# -*- coding: utf-8 -*-
from typing import List, Tuple

from matplotlib.axes import Axes
from matplotlib.figure import Figure
from mpl_toolkits.mplot3d.axes3d import Axes3D

from ..borehole.borehole import Borehole
from ..borehole.single_u_tube import SingleUTube
from .plot_path import plot_path


def plot_borehole(borehole: Borehole | SingleUTube, num: int = 101, view: str = 'all', ax: Axes | Axes3D | None = None) -> Axes | Axes3D | Tuple[Figure, Axes] | Tuple[Figure, Axes3D] | Tuple[Figure, List[Axes | Axes3D]]:
    """Plot borehole trajectory

    Parameters
    ----------
    borehole : borehole or single_u_tube
        Path to be plotted.
    num : int, default: 101
        Number of points along the borehole.
    view : {'all', 'xy', 'yz', 'xz', '3d'}, default: 'all'
        The view to be plotted. 'all' plots all views on a 2 by 2
        figure.
    ax : axes, axes_3d or None, default: ``None``
        The axis on which to draw the trajectories. Is `ax` is ``None``,
        a new figure and new axes are created.

    Returns
    -------
    fig : figure
        The figure. Only return if `ax` is ``None``.
    ax : axes, axes_3d or list
        The axis.

    """
    return plot_path(borehole.path, num=num, view=view, ax=ax)



# -*- coding: utf-8 -*-
from collections.abc import Callable
from functools import partial
from typing import Self, Tuple

from jax import numpy as jnp
from jax import Array, jit, vmap
from jax.typing import ArrayLike

from ._tube import _Tube
from ..basis import Basis
from ..path import Path


class SingleUTube(_Tube):
    """Single U-Tube geothermal borehole.

    Parameters
    ----------
    R_d : array_like or callable
        (2, 2,) array of thermal resistances (in m-K/W), or callable that
        takes the mass flow rate as input (in kg/s) and returns a (2, 2,)
        array.
    r_b : float
        Borehole radius (in meters).
    path : path
        Path of the borehole.
    basis : basis
        Basis functions.
    n_segments : int
        Number of segments.
    segment_ratios : array_like or None, default: None
        Normalized size of the segments. Should total ``1``
        (i.e. ``sum(segment_ratios) = 1``). If `segment_ratios` is
        ``None``, segments of equal size are considered (i.e.
        ``segment_ratios[v] = 1 / n_segments``).

    Attributes
    ----------
    n_pipes : int
        Number of pipes in the borehole.
    n_nodes : int
        Total number of nodes along the borehole.
    xi_edges : array
        (`n_segments`+1,) array of the coordinates of the edges of the
        segments.
    L : float
        Borehole length (in meters).
    xi : array
        (`n_nodes`,) array of node coordinates.
    p : array
        (`n_nodes`, 3,) array of node positions.
    dp_dxi : array
        (`n_nodes`, 3,) array of the derivatives of the position at the
        node coordinates.
    J : array
        (`n_nodes`,) array of the norm of the Jacobian at the node
        coordinates.
    s : array
        (`n_nodes`,) array of the longitudinal position at the node
        coordinates.
    w : array
        (`n_nodes`,) array of quadrature weights at the node coordinates.
        These quadrature weights take into account the norm of the
        Jacobian.

    """

    def _fluid_temperature_a_in(self, xi: Array | float, beta_ij: Array) -> Array:
        """Inlet coefficient to evaluate the fluid temperatures.

        Parameters
        ----------
        xi : array or float
            (M,) array of coordinates along the borehole.
        beta_ij: array
            (`n_pipes`, `n_pipes`,) array of thermal conductance
            coefficients.

        Returns
        -------
        array
            (M, `n_pipes`,) array of coefficients for the inlet fluid
            temperature.

        """
        b_in = self._outlet_fluid_temperature_a_in(beta_ij)
        c_in = self._general_solution_a_in(xi, beta_ij)
        c_out = self._general_solution_a_out(xi, beta_ij)
        a_in = c_in + b_in * c_out
        return a_in

    def _fluid_temperature_a_b(self, xi: Array | float, beta_ij: Array) -> Array:
        """Borehole wall coefficient to evaluate the fluid temperatures.

        Parameters
        ----------
        xi : array or float
            (M,) array of coordinates along the borehole.
        beta_ij: array
            (`n_pipes`, `n_pipes`,) array of thermal conductance
            coefficients.

        Returns
        -------
        array
            (M, `n_nodes`,) array of coefficients for the borehole wall
            temperature.

        """
        b_b = self._outlet_fluid_temperature_a_b(beta_ij)
        c_out = self._general_solution_a_out(xi, beta_ij)
        c_b = self._general_solution_a_b(xi, beta_ij)
        a_b = c_b + vmap(jnp.outer, in_axes=(0, None))(c_out, b_b)
        return a_b

    def _general_solution(self, xi: Array | float, m_flow: float, cp_f: float) -> Tuple[Array, Array, Array]:
        """Coefficients to evaluate the general solution.

        Parameters
        ----------
        xi : array or float
            (M,) array of coordinates along the borehole.
        m_flow : float
            Fluid mass flow rate (in kg/s).
        cp_f : float
            Fluid specific isobaric heat capacity (in J/kg-K).

        Returns
        -------
        a_in : array or float
            (M, `n_pipes`,) array of coefficients for the inlet fluid
            temperature.
        a_out : array or float
            (M, `n_pipes`,) array of coefficients for the outlet fluid
            temperature.
        a_b : array
            (M, `n_pipes`, `n_nodes`,) array of coefficients for the
            borehole wall temperature.

        """
        beta_ij = self._beta_ij(m_flow, cp_f)
        a_in = self._general_solution_a_in(xi, beta_ij)
        a_out = self._general_solution_a_out(xi, beta_ij)
        a_b = self._general_solution_a_b(xi, beta_ij)
        return a_in, a_out, a_b

    def _general_solution_a_in(self, xi: Array | float, beta_ij: Array) -> Array:
        """Inlet coefficient to evaluate the general solution.

        Parameters
        ----------
        xi : array or float
            (M,) array of coordinates along the borehole.
        beta_ij: array
            (2, 2,) array of thermal conductance coefficients.

        Returns
        -------
        array or float
            (M, 2,) array of coefficients for the inlet fluid temperature.

        """
        s = self.path.f_s(xi)
        a_in = jnp.stack(
            (
                self._f1(s, beta_ij),
                -self._f2(s, beta_ij)
            ),
            axis=-1
        )
        return a_in

    def _general_solution_a_out(self, xi: Array | float, beta_ij: Array) -> Array:
        """Outlet coefficient to evaluate the general solution.

        Parameters
        ----------
        xi : array or float
            (M,) array of coordinates along the borehole.
        beta_ij: array
            (2, 2,) array of thermal conductance coefficients.

        Returns
        -------
        array or float
            (M, 2,) array of coefficients for the outlet fluid
            temperature.

        """
        s = self.path.f_s(xi)
        a_out = jnp.stack(
            (
                self._f2(s, beta_ij),
                self._f3(s, beta_ij)
            ),
            axis=-1
        )
        return a_out

    def _general_solution_a_b(self, xi: Array | float, beta_ij: Array) -> Array:
        """Borehole wall coefficient to evaluate the general solution.

        Parameters
        ----------
        xi : array or float
            (M,) array of coordinates along the borehole.
        beta_ij: array
            (2, 2,) array of thermal conductance coefficients.

        Returns
        -------
        array
            (M, 2, `n_nodes`,) array of coefficients for the borehole wall
            temperature.

        """
        if len(jnp.shape(xi)) == 1:
            a_b = vmap(
                self._general_solution_a_b,
                in_axes=(0, None)
            )(xi, beta_ij)
        else:
            s = self.path.f_s(xi)
            f1 = lambda _eta: self._f4(s - self.path.f_s(_eta), beta_ij) * self.path.f_J(_eta)
            f2 = lambda _eta: -self._f5(s - self.path.f_s(_eta), beta_ij) * self.path.f_J(_eta)
            high = jnp.maximum(-1., jnp.minimum(1., self.f_xi_bs(xi)))
            a, b = self.xi_edges[:-1], self.xi_edges[1:]
            f_xi_bs = lambda _eta, _a, _b: 0.5 * (_b + _a) + 0.5 * _eta * (_b - _a)
            integrand = lambda _eta, _a, _b, _ratio: jnp.stack(
                [
                    f1(f_xi_bs(_eta, _a, _b)) * _ratio,
                    f2(f_xi_bs(_eta, _a, _b)) * _ratio,
                ]
            )
            integral = lambda _a, _b, _ratio, _high: self.basis.quad_gl(
                    vmap(
                        lambda _eta: integrand(_eta, _a, _b, _ratio),
                        in_axes=0,
                        out_axes=-1
                    ),
                -1.,
                _high
                )
            a_b = vmap(
                    integral,
                    in_axes=(0, 0, 0, 0),
                    out_axes=1
                )(a, b, self.segment_ratios, high).reshape(2, self.n_nodes)
        return a_b

    def _outlet_fluid_temperature_a_in(self, beta_ij: Array) -> float:
        """Inlet coefficient to evaluate the outlet fluid temperature.

        Parameters
        ----------
        beta_ij: array
            (2, 2,) array of thermal conductance coefficients.

        Returns
        -------
        float
            Coefficient for the inlet fluid temperature.

        """
        L = self.L
        a_in = (self._f1(L, beta_ij) + self._f2(L, beta_ij)) / (self._f3(L, beta_ij) - self._f2(L, beta_ij))
        return a_in

    def _outlet_fluid_temperature_a_b(self, beta_ij: Array) -> Array:
        """Borehole coefficient to evaluate the outlet fluid temperature.

        Parameters
        ----------
        beta_ij: array
            (2, 2,) array of thermal conductance coefficients.

        Returns
        -------
        array
            (`n_nodes`,) array of coefficients for the borehole wall
            temperature.

        """
        L = self.L
        f = vmap(
            lambda _eta: (
                self._f4(L - self.path.f_s(self.f_xi_sb(_eta)), beta_ij)
                + self._f5(L - self.path.f_s(self.f_xi_sb(_eta)), beta_ij)
            ) / (
                self._f3(L, beta_ij)
                - self._f2(L, beta_ij)
            ) * self.path.f_J(self.f_xi_sb(_eta)) * self.segment_ratios,
            in_axes=0,
            out_axes=-1)
        a_b = self.basis.quad_gl(f, -1, 1.).flatten()
        return a_b

    def _beta(self, beta_ij: Array) -> float:
        """Coefficient ``beta`` from Hellström (1991).

        Parameters
        ----------
        beta_ij: array
            (2, 2,) array of thermal conductance coefficients.

        Returns
        -------
        float

        """
        beta1 = beta_ij[0, 0]
        beta2 = beta_ij[1, 1]
        beta = 0.5 * (beta2 - beta1)
        return beta

    def _gamma(self, beta_ij: Array) -> float:
        """Coefficient ``gamma`` from Hellström (1991).

        Parameters
        ----------
        beta_ij: array
            (2, 2,) array of thermal conductance coefficients.

        Returns
        -------
        float

        """
        beta1 = beta_ij[0, 0]
        beta2 = beta_ij[1, 1]
        beta12 = beta_ij[0, 1]
        gamma = jnp.sqrt(0.25 * (beta1 + beta2)**2 + beta12 * (beta1 + beta2))
        return gamma

    def _delta(self, beta_ij: Array) -> float:
        """Coefficient ``delta`` from Hellström (1991).

        Parameters
        ----------
        beta_ij: array
            (2, 2,) array of thermal conductance coefficients.

        Returns
        -------
        float

        """
        beta1 = beta_ij[0, 0]
        beta2 = beta_ij[1, 1]
        beta12 = beta_ij[0, 1]
        gamma = self._gamma(beta_ij)
        delta = 1. / gamma * (beta12 + 0.5 * (beta1 + beta2))
        return delta

    def _f1(self, s: Array | float, beta_ij: Array) -> Array | float:
        """Function ``f1`` from Hellström (1991).

        Parameters
        ----------
        s : array or float
            (M,) array of longitudinal positions (in meters).
        beta_ij: array
            (2, 2,) array of thermal conductance coefficients.

        Returns
        -------
        array or float
            (M,) array.

        """
        beta = self._beta(beta_ij)
        gamma = self._gamma(beta_ij)
        delta = self._delta(beta_ij)
        f1 = jnp.exp(beta * s) * (jnp.cosh(gamma * s) - delta * jnp.sinh(gamma * s))
        return f1

    def _f2(self, s: Array | float, beta_ij: Array) -> Array | float:
        """Function ``f2`` from Hellström (1991).

        Parameters
        ----------
        s : array or float
            (M,) array of longitudinal positions (in meters).
        beta_ij: array
            (2, 2,) array of thermal conductance coefficients.

        Returns
        -------
        array or float
            (M,) array.

        """
        beta = self._beta(beta_ij)
        gamma = self._gamma(beta_ij)
        beta12 = beta_ij[0, 1]
        f2 = jnp.exp(beta * s) * beta12 / gamma * jnp.sinh(gamma * s)
        return f2

    def _f3(self, s: Array | float, beta_ij: Array) -> Array | float:
        """Function ``f3`` from Hellström (1991).

        Parameters
        ----------
        s : array or float
            (M,) array of longitudinal positions (in meters).
        beta_ij: array
            (2, 2,) array of thermal conductance coefficients.

        Returns
        -------
        array or float
            (M,) array.

        """
        beta = self._beta(beta_ij)
        gamma = self._gamma(beta_ij)
        delta = self._delta(beta_ij)
        f3 = jnp.exp(beta * s) * (jnp.cosh(gamma * s) + delta * jnp.sinh(gamma * s))
        return f3

    def _f4(self, s: Array | float, beta_ij: Array) -> Array | float:
        """Function ``f4`` from Hellström (1991).

        Parameters
        ----------
        s : array or float
            (M,) array of longitudinal positions (in meters).
        beta_ij: array
            (2, 2,) array of thermal conductance coefficients.

        Returns
        -------
        array or float
            (M,) array.

        """
        beta1 = beta_ij[0, 0]
        beta2 = beta_ij[1, 1]
        beta12 = beta_ij[0, 1]
        beta = self._beta(beta_ij)
        gamma = self._gamma(beta_ij)
        delta = self._delta(beta_ij)
        f4 = jnp.exp(beta * s) * (beta1 * jnp.cosh(gamma * s) - (delta * beta1 + beta2 * beta12 / gamma) * jnp.sinh(gamma * s))
        return f4

    def _f5(self, s: Array | float, beta_ij: Array) -> Array | float:
        """Function ``f5`` from Hellström (1991).

        Parameters
        ----------
        s : array or float
            (M,) array of longitudinal positions (in meters).
        beta_ij: array
            (2, 2,) array of thermal conductance coefficients.

        Returns
        -------
        array or float
            (M,) array.

        """
        beta1 = beta_ij[0, 0]
        beta2 = beta_ij[1, 1]
        beta12 = beta_ij[0, 1]
        beta = self._beta(beta_ij)
        gamma = self._gamma(beta_ij)
        delta = self._delta(beta_ij)
        f5 = jnp.exp(beta * s) * (beta2 * jnp.cosh(gamma * s) + (delta * beta2 + beta1 * beta12 / gamma) * jnp.sinh(gamma * s))
        return f5



# -*- coding: utf-8 -*-
from functools import partial

from jax import numpy as jnp
from jax import Array, jit, vmap
from jax.lax import fori_loop
from jax.typing import ArrayLike

from ..utilities import comb


class Multipole:
    """Multipole method for 3d heat transfer.

    Parameters
    ----------
    r_b : float
        Borehole radius (in meters).
    r_p : array_like
        Outer radius of the pipes (in meters).
    p : array_like
        (`n_pipes`, 2) array of pipe positions (in meters). The position
        ``(0, 0)`` corresponds to the axis of the borehole.
    k_s : float
        Ground thermal conductivity (in W/m-K).
    k_b : float
        Grout thermal conductivity (in W/m-K).
    J : int, default: 3
        Order of the multipole solution.

    Attributes
    ----------
    n_pipes : int
        Number of pipes.
    sigma : float
        Dimensionless parameter for the two thermal conductivities.
    z : complex array
        (`n_pipes`,) complex array of pipe positions (in meters).

    """

    def __init__(self, r_b: float, r_p: ArrayLike, p: ArrayLike, k_s: float, k_b: float, J: int = 3):
        # Runtime type validation
        if not isinstance(r_p, ArrayLike):
            raise TypeError(f"Expected arraylike input; got {r_p}")
        if not isinstance(p, ArrayLike):
            raise TypeError(f"Expected arraylike input; got {p}")
        # Convert input to jax.Array
        r_p = jnp.asarray(r_p)
        p = jnp.atleast_2d(p)

        # --- Class atributes ---
        # Parameters
        self.r_b = r_b
        self.r_p = r_p
        self.p = p
        self.k_s = k_s
        self.k_b = k_b
        self.J = J
        # Additional attributes
        self.n_pipes = len(r_p)
        self.sigma = (k_b - k_s) / (k_b + k_s)
        self.z = p[:, 0] + 1.j * p[:, 1]
        self._zz = vmap(
            jnp.multiply,
            in_axes=(0, None),
            out_axes=0
            )(self.z.conj(), self.z)
        self._dis = vmap(
            jnp.add,
            in_axes=(0, None),
            out_axes=0
            )(self.z, -self.z)

        # Initialize mutipole coefficients and resistances
        self._initialize_factors()
        self._initialize_thermal_resistances()
        self._initialize_system_of_equations()

    def beta(self, R_fp: Array) -> Array:
        """Dimensionless fluid to pipe thermal resistances.

        Parameters
        ----------
        R_fp : array
            (`n_pipes`,) array of fluid to pipe thermal resistance
            (in m-K/W).

        Returns
        -------
        array
            Dimensionless fluid to pipe thermal resistances.

        """
        return R_fp * (2 * jnp.pi * self.k_b)

    @partial(jit, static_argnames=['self'])
    def fluid_temperatures(self, q: Array, T_b: float, R_fp: Array) -> Array:
        """Evaluate fluid temperatures.

        Parameters
        ----------
        q : array
            (`n_pipes`,) array of heat injection rates in each pipe
            (in W/m).
        T_b : float
            Average borehole wall temperature (in degree Celsius).
        R_fp : array
            (`n_pipes`,) array of fluid to pipe thermal resistance
            (in m-K/W).

        Returns
        -------
        array
            (`n_pipes`,) array of Fluid temperatures (in degree Celsius).

        """
        # Dimensionless fluid to pipe thermal resistances
        beta = self.beta(R_fp)
        # Multipole strengths (eqs. 34-36)
        P = self.multipoles(q, beta)
        # Fluid temperatures (eq. 32)
        a_q = self._thermal_resistances_line(beta)
        a_P = self.thermal_resistances_multipole
        T_f = T_b + a_q @ q + jnp.tensordot(a_P, P).real
        return T_f

    @partial(jit, static_argnames=['self'])
    def multipoles(self, q: Array, beta: Array) -> Array:
        """Evaluate multipoles.

        Parameters
        ----------
        q : array
            (`n_pipes`,) array of heat injection rates in each pipe
            (in W/m).
        beta : array
            Dimensionless fluid to pipe thermal resistances.

        Returns
        -------
        array
            (`n_pipes`, `J`,) array of multipole strengths.

        """
        N = self.n_pipes * self.J
        k = jnp.arange(1, self.J + 1)
        # Coefficients that depend on the fluid to pipe thermal
        # resistances (eq. 36)
        beta_k = vmap(
            jnp.multiply,
            in_axes=(0, None),
            out_axes=0
            )(beta, k)
        coeffs = (1 + beta_k) / (1 - beta_k)
        diag_indices = jnp.diag_indices_from(self.A)
        A = self.A.at[*diag_indices].add(jnp.concatenate([coeffs.flatten(), -coeffs.flatten()]))
        # Right hand side (eq. 36)
        B = (-self.factors_line @ q).flatten()
        B = jnp.concatenate([B.real, B.imag])
        # Solve for multipoles (eq. 36)
        P = jnp.linalg.solve(A, B)
        P = (P[:N] + 1.j * P[N:]).reshape((self.n_pipes, self.J))
        return P

    @partial(jit, static_argnames=['self'])
    def thermal_resistances(self, R_fp: Array) -> Array:
        """Evaluate delta-circuit thermal resistances.

        Parameters
        ----------
        R_fp : array
            (`n_pipes`,) array of fluid to pipe thermal resistance
            (in m-K/W).

        Returns
        -------
        array
            (`n_pipes`, `n_pipes`,) array of delta-circuit thermal
            resistances (in m-K/W).

        """
        # Evaluate fluid temperatures for unit heat transfer rates at
        # each pipe, individually (eq. 32).
        R = vmap(
            self.fluid_temperatures,
            in_axes=(0, None, None),
            out_axes=0
            )(jnp.eye(self.n_pipes), 0., R_fp)
        # Thermal conductances (eq. 51)
        K = -jnp.linalg.inv(R)
        # Delta-circuit thermal resistances (eq. 51)
        diag_indices = jnp.diag_indices_from(K)
        K = K.at[*diag_indices].set(-K.sum(axis=1))
        R_d = 1 / K
        return R_d

    def _factors_line(self, k: int) -> Array:
        """Multipole factor coefficients for the heat injection rates.

        Parameters
        ----------
        k : int
            Target multipole order.

        Returns
        -------
        array
            (`n_pipes`, `n_pipes`,) array of coefficients

        """
        z = self.z
        zz = self._zz
        I = jnp.eye(self.n_pipes)
        # First term (eq. 34)
        denom_1 = ((1 - I) * -self._dis + I)**k
        numer_1 = ((1 - I) * self.r_p**k).T
        # Second term (eq. 34)
        denom_2 = (self.r_b**2 - zz.T)**k
        numer_2 = self.sigma * vmap(
            jnp.multiply,
            in_axes=(0, None),
            out_axes=0
            )(self.r_p**k, z.conj()**k)
        # Sum of the two terms
        a_q = (numer_1 / denom_1 + numer_2 / denom_2) / k / (2 * jnp.pi * self.k_b)
        return a_q

    def _factors_multipole(self, k: int, j: int) -> Array:
        """Multipole factor coefficients for the multipoles.

        Parameters
        ----------
        k : int
            Target multipole order.
        J : int
            Source multipole order.

        Returns
        -------
        array
            (`n_pipes`, `n_pipes`,) array of coefficients

        """
        I = jnp.eye(self.n_pipes)
        # Third term (eq. 34)
        numerator = vmap(
            jnp.multiply,
            in_axes=(0, None),
            out_axes=0
            )((-self.r_p)**k, self.r_p**j)
        denominator = self._dis**(k + j) + I
        a_P = (1 - I) * comb(k + j - 1, j - 1) * numerator / denominator
        return a_P

    def _factors_conjugate_multipole(self, k: int, j: int) -> Array:
        """Multipole factor coefficients for the conjugate multipoles.

        Parameters
        ----------
        k : int
            Target multipole order.
        J : int
            Source multipole order.

        Returns
        -------
        array
            (`n_pipes`, `n_pipes`,) array of coefficients

        """
        j_p_max = jnp.minimum(k, j)
        z = self.z
        # Fourth term (eq. 34)
        # This is a for loop but uses jax.lax.fori_loop to allow tracing.
        a_P_conj = jnp.zeros((self.n_pipes, self.n_pipes), dtype=complex)
        def body_fun(i, val):
            numerator = vmap(
                jnp.multiply,
                in_axes=(0, None),
                out_axes=0
                )(self.r_p**k * z**(j - i), self.r_p**j * z.conj()**(k - i))
            denominator = (self.r_b**2 - self._zz.T)**(k + j - i)
            return val + self.sigma * comb(j, i) * comb(k + j - i - 1, j - 1) * numerator / denominator
        a_P_conj = fori_loop(0, j_p_max + 1, body_fun, a_P_conj)
        return a_P_conj

    def _initialize_factors(self):
        """Initialize multipole factor coefficients.

        """
        k = jnp.arange(1, self.J + 1)
        j = jnp.arange(1, self.J + 1)
        self.factors_line = vmap(
            self._factors_line,
            in_axes=0,
            out_axes=1
        )(k)
        self.factors_multipole = vmap(
            vmap(
                self._factors_multipole,
                in_axes=(0, None),
                out_axes=1
                ),
            in_axes=(None, 0),
            out_axes=-1
        )(k, j)
        self.factors_conjugate_multipole = vmap(
            vmap(
                self._factors_conjugate_multipole,
                in_axes=(0, None),
                out_axes=1
                ),
            in_axes=(None, 0),
            out_axes=-1
        )(k, j)

    def _initialize_system_of_equations(self):
        """Initialize the system of equations for multipoles.

        """
        N = self.n_pipes * self.J
        A = self.factors_multipole.reshape((N, N))
        C = self.factors_conjugate_multipole.reshape((N, N))
        # Split the real and image parts to allow the solution of a linear
        # system of equations (eq. 36). The coefficients `C` multiply the
        # conjugate of the multipoles.
        self.A = jnp.block(
            [[A.real + C.real, -A.imag + C.imag],
             [A.imag + C.imag, A.real - C.real]]
            )

    def _initialize_thermal_resistances(self):
        """Initialize thermal resistances.

        """
        self.thermal_resistances_line_zero_beta = (
            self._thermal_resistances_line_zero_beta()
            )
        self.thermal_resistances_multipole = vmap(
            self._thermal_resistances_multipole,
            in_axes=0,
            out_axes=-1
            )(jnp.arange(1, self.J + 1))

    def _thermal_resistances_line(self, beta: Array) -> Array:
        """Evaluate thermal resistances for heat extraction rates.

        Parameters
        ----------
        beta : array
            Dimensionless fluid to pipe thermal resistances.

        Returns
        -------
        array
            (`n_pipes`, `n_pipes`,) array of thermal resistances
            (in m-K/W).

        """
        diag_indices = jnp.diag_indices_from(
            self.thermal_resistances_line_zero_beta
            )
        R = self.thermal_resistances_line_zero_beta.at[*diag_indices].add(
            beta / (2 * jnp.pi * self.k_b)
            )
        return R

    def _thermal_resistances_line_zero_beta(self) -> Array:
        """Initialize thermal resistances for heat extraction rates.

        Returns
        -------
        array
            (`n_pipes`, `n_pipes`,) array of thermal resistances
            (in m-K/W).

        """
        z = self.z
        zz = self._zz
        I = jnp.eye(self.n_pipes)
        absdis = jnp.abs(self._dis)
        # Thermal resistances (eq. 33), assuming `beta=0`
        R = 1 / (2 * jnp.pi * self.k_b) * (
            jnp.log(self.r_b / ((1 - I) * absdis + jnp.diag(self.r_p)))
            + self.sigma * jnp.log(
                self.r_b**2 / (
                    (1 - I) * jnp.abs(self.r_b**2 - zz)
                    + jnp.diag(self.r_b**2 - jnp.abs(z)**2)
                )
            )
        )
        return R

    def _thermal_resistances_multipole(self, j: int) -> Array:
        """Evaluate thermal resistances for multipoles.

        Parameters
        ----------
        j : int
            Multipole order.

        Returns
        -------
        array
            (`n_pipes`, `n_pipes`,) array of thermal resistances.

        """
        z = self.z
        zz = self._zz
        I = jnp.eye(self.n_pipes)
        # Thermal resistances of multipoles (eq. 32)
        numer_1 = (1 - I) * self.r_p**j
        denom_1 = self._dis**j + I
        numer_2 = vmap(
            jnp.multiply,
            in_axes=(0, None),
            out_axes=0
            )(z.conj()**j, self.r_p**j)
        denom_2 = (self.r_b**2 - zz)**j
        a_P = (numer_1 / denom_1 + self.sigma * numer_2 / denom_2)
        return a_P



# -*- coding: utf-8 -*-
from jax import jit
from jax import numpy as jnp
from jax.scipy.special import gammaln


@jit
def comb(N: int, k: int) -> float:
    """Number of combinations (N choose k).

    Parameters
    ----------
    N : int
        Number of things.
    k : int
        Number of elements taken.

    Returns
    -------
    float
        The total number of combinations.

    """
    return jnp.exp(gammaln(N + 1) - gammaln(k + 1) - gammaln(N - k + 1))



# -*- coding: utf-8 -*-
from typing import List

from matplotlib import pyplot as plt
from matplotlib.axes import Axes
from matplotlib.ticker import AutoMinorLocator
from mpl_toolkits.mplot3d.axes3d import Axes3D


def _format_axis(ax: Axes | Axes3D, axis_labels: List[str] | None = None, equal: bool = False, inverse_y: bool = False) -> Axes | Axes3D:
    """Format an axis.

    Parameters
    ----------
    ax : axes or axes_3d
        Axis to be formatted.
    axis_labels : list of str or None, default: ``None``
        Axis labels.
    equal : bool, default: ``False``
        True to set use an equal aspect ratio.
    inverse_y : bool, default: ``False``
        True to invert the direction of the y axis.

    Returns
    -------
    axes or axes_3d
        The formatted axis.
    """
    ax.tick_params(
        axis='both', which='both', direction='in',
        bottom=True, top=True, left=True, right=True)
    ax.xaxis.set_minor_locator(AutoMinorLocator())
    ax.yaxis.set_minor_locator(AutoMinorLocator())
    if axis_labels is not None:
        ax.set_xlabel(axis_labels[0])
        ax.set_ylabel(axis_labels[1])
        if ax.name == '3d' and len(axis_labels) == 3:
            ax.set_zlabel(axis_labels[2])
    if isinstance(ax, Axes3D):
        ax.zaxis.set_minor_locator(AutoMinorLocator())
    if equal:
        ax.set_aspect('equal', adjustable='datalim')
    if inverse_y:
        ax.yaxis.set_inverted(True)
    return ax



# -*- coding: utf-8 -*-
from jax import numpy as jnp


def darcy_friction_factor(Re: float, E: float) -> float:
    """
    Darcy-Weisbach friction factor.

    Parameters
    ----------
    Re : float
        Reynolds number.
    E : float
        Relative pipe roughness (epsilon / D).

    Returns
    -------
    f_darcy : float
        Darcy friction factor.

    """
    Re_crit = 2300.
    # Condition on the flow regime
    condlist = [Re <= Re_crit]
    # Piecewise evaluation of Darcy friction factor
    funclist = [
        darcy_friction_factor_laminar_flow,
        darcy_friction_factor_turbulent_flow
    ]
    f_darcy = jnp.piecewise(jnp.float32(Re), condlist, funclist, E)
    return f_darcy


def darcy_friction_factor_laminar_flow(Re: float, *args) -> float:
    """
    Darcy-Weisbach friction factor for laminar flow in circular pipe.

    Parameters
    ----------
    Re : float
        Reynolds number.

    Returns
    -------
    f_darcy : float
        Darcy friction factor.

    """
    return 64 / Re


def darcy_friction_factor_turbulent_flow(Re: float, E: float) -> float:
    """
    Darcy-Weisbach friction factor for turbulent flow in circular pipe.

    Parameters
    ----------
    Re : float
        Reynolds number.
    E : float
        Relative pipe roughness (epsilon / D).

    Returns
    -------
    f_darcy : float
        Darcy friction factor.

    """
    # Churchill equation for rough pipes
    A = (
        2.457 * jnp.log(
            1 / ((7 / Re)**0.9 + (0.27 * E))
            )
        )**16
    B = (37_530 / Re)**16
    f_darcy = 8 * ((8 / Re)**12 + 1 / (A + B)**1.5)**(1 / 12)
    return f_darcy



# -*- coding: utf-8 -*-
from functools import partial
from typing import Self

from jax import numpy as jnp
from jax import Array, jit, vmap
from jax.typing import ArrayLike

from ..basis import Basis
from ..path import Path


class Borehole:
    """Geothermal borehole.

    Parameters
    ----------
    r_b : float
        Borehole radius (in meters).
    path : path
        Path of the borehole.
    basis : basis
        Basis functions.
    n_segments : int
        Number of segments.
    segment_ratios : array_like or None, default: None
        Normalized size of the segments. Should total ``1``
        (i.e. ``sum(segment_ratios) = 1``). If `segment_ratios` is
        ``None``, segments of equal size are considered (i.e.
        ``segment_ratios[v] = 1 / n_segments``).

    Attributes
    ----------
    n_nodes : int
        Total number of nodes along the borehole.
    xi_edges : array
        (`n_segments`+1,) array of the coordinates of the edges of the
        segments.
    L : float
        Borehole length (in meters).
    xi : array
        (`n_nodes`,) array of node coordinates.
    p : array
        (`n_nodes`, 3,) array of node positions.
    dp_dxi : array
        (`n_nodes`, 3,) array of the derivatives of the position at the
        node coordinates.
    J : array
        (`n_nodes`,) array of the norm of the Jacobian at the node
        coordinates.
    s : array
        (`n_nodes`,) array of the longitudinal position at the node
        coordinates.
    w : array
        (`n_nodes`,) array of quadrature weights at the node coordinates.
        These quadrature weights take into account the norm of the
        Jacobian.

    """

    def __init__(self, r_b: float, path: Path, basis: Basis, n_segments: int, segment_ratios: ArrayLike | None = None):
        # Runtime type validation
        if not isinstance(segment_ratios, ArrayLike) and segment_ratios is not None:
            raise TypeError(f"Expected arraylike or None input; got {segment_ratios}")
        # Convert input to jax.Array
        if segment_ratios is None:
            segment_ratios = jnp.full(n_segments, 1. / n_segments)
        else:
            segment_ratios = jnp.asarray(segment_ratios)

        # --- Class attributes ---
        self.r_b = r_b
        self.path = path
        self.basis = basis
        self.n_segments = n_segments
        self.n_nodes = basis.n_nodes * n_segments
        self.segment_ratios = segment_ratios
        # Segment edges
        xi_edges = jnp.concatenate((-jnp.ones(1), 2 * jnp.cumsum(segment_ratios) - 1))
        self.xi_edges = xi_edges
        # Borehole length
        self.L = jnp.diff(path.f_s(jnp.array([-1., 1.])))[0]

        # --- Changes of coordinates ---
        a, b = xi_edges[:-1], xi_edges[1:]
        # Segments --> Borehole
        f_xi_sb = lambda _eta: 0.5 * (b + a) + 0.5 * _eta * (b - a)
        self.f_xi_sb = jit(
            lambda _eta: vmap(f_xi_sb, in_axes=0)(_eta) if len(jnp.shape(_eta)) > 0 else f_xi_sb(_eta)
        )
        # Borehole --> Segments
        f_xi_bs = lambda _eta: (2 * _eta - (b + a)) / (b - a)
        self.f_xi_bs = jit(
            lambda _eta: vmap(f_xi_bs, in_axes=0)(_eta) if len(jnp.shape(_eta)) > 0 else f_xi_bs(_eta)
        )

        # --- Basis functions ---
        # Weights at interfaces between segments
        w_interfaces = jnp.concatenate([jnp.array([1.]), jnp.full(n_segments - 1, 0.5), jnp.array([1.])])
        f_w_interfaces = lambda _eta: jnp.heaviside(b - _eta, w_interfaces[1:]) * jnp.heaviside(_eta - a, w_interfaces[:-1])
        f_psi = lambda _eta: vmap(
            lambda _eta_p, _in_segment: basis.f_psi(_eta_p) * _in_segment,
            in_axes=(0, 0)
        )(f_xi_bs(_eta), f_w_interfaces(_eta)).flatten()
        self.f_psi = jit(
            lambda _eta: vmap(f_psi, in_axes=0)(_eta) if len(jnp.shape(_eta)) > 0 else f_psi(_eta)
        )
        self.f = jit(
            lambda _eta, f_nodes: self.f_psi(_eta) @ f_nodes
        )

        # --- Nodal values of path and basis functions ---
        # Borehole coordinates (xi)
        xi = self.f_xi_sb(basis.xi).T.flatten()
        self.xi = xi
        # Positions (p)
        self.p = path.f_p(xi)
        # Derivatives of position (dp/dxi)
        self.dp_dxi = path.f_dp_dxi(xi)
        # Norms of the Jacobian (J)
        self.J = path.f_J(xi)
        # Longitudinal positions (s)
        self.s = path.f_s(xi)
        # Integration weights
        self.w = (jnp.tile(basis.w, (n_segments, 1)).T * segment_ratios).T.flatten() * self.J

    def h_to_borehole(self, borehole: Self, time: ArrayLike, alpha: float) -> Array:
        """Thermal response factors to nodes of another borehole.

        Parameters
        ----------
        borehole : borehole
            Borehole for which the thermal response factors will be
            evaluated at the nodes.
        time : array
            (K,) array of times (in seconds).
        alpha : float
            Ground thermal diffusivity (in m^2/s).

        Returns
        -------
        array
            (K, ```borehole`.n_nodes``, `n_nodes`,) array of thermal
            response factors.
        """
        return self.h_to_point(borehole.p, time, alpha)

    @partial(jit, static_argnames=['self'])
    def h_to_coordinate_on_self(self, xi: Array, time: Array, alpha: float) -> Array:
        """Thermal response factors to coordinates along itself.

        Parameters
        ----------
        xi : array
            (M,) array of the coordinates along itself at which thermal
            response factors are evaluated.
        time : array
            (K,) array of times (in seconds).
        alpha : float
            Ground thermal diffusivity (in m^2/s).

        Returns
        -------
        array
            (K, M, `n_nodes`,) array of thermal response factors.
        """
        # Positions (p) of points on self
        p = self.path.f_p(xi)
        return self.h_to_point(p, time, alpha, r_min=self.r_b)

    @partial(jit, static_argnames=['self'])
    def h_to_point(self, p: Array, time: Array, alpha: float, r_min: float = 0.):
        """Thermal response factors to a point.

        Parameters
        ----------
        p : array
            (M, 3,) array of the positions at which thermal response
            factors are evaluated.
        time : array
            (K,) array of times (in seconds).
        alpha : float
            Ground thermal diffusivity (in m^2/s).
        r_min : float, default: ``0.``
            Minimum distance (in meters) added to the distance between the
            heat source and the positions `p`.

        Returns
        -------
        array
            (K, M, `n_nodes`,) array of thermal response factors.
        """
        # Integrand of point heat source
        integrand = vmap(
            lambda _eta: self._segment_point_heat_source(_eta, p, time, alpha, r_min=r_min),
            in_axes=0,
            out_axes=-1)
        n_times = len(time)
        n_nodes = self.n_nodes
        h_to_point = self.basis.quad_gl(
            integrand, -1., 1.
        ).reshape(n_times, -1, n_nodes)
        return h_to_point

    @partial(jit, static_argnames=['self'])
    def h_to_self(self, time: Array, alpha: float):
        """Thermal response factors to its own nodes.

        Parameters
        ----------
        time : array
            (K,) array of times (in seconds).
        alpha : float
            Ground thermal diffusivity (in m^2/s).

        Returns
        -------
        array
            (K, `n_nodes`, `n_nodes`,) array of thermal response factors.
        """
        # Integrand of point heat source evaluated at borehole nodes
        integrand = vmap(
            lambda _eta: self._segment_point_heat_source(_eta, self.p, time, alpha, r_min=self.r_b),
            in_axes=0,
            out_axes=-1)
        # Integral of the point heat source
        n_times = len(time)
        n_nodes = self.n_nodes
        h_to_self = self.basis.quad_ts_nodes(
            integrand
        ).reshape(n_times, n_nodes, n_nodes)
        return h_to_self

    @partial(jit, static_argnames=['self'])
    def _segment_point_heat_source(self, xi_p: float, p: Array, time: Array | float, alpha: float, r_min: float = 0.) -> Array:
        """Point heat source solution along all segments of the borehole.

        Parameters
        ----------
        xi_p : array
            (N,) array of coordinates along segments.
        p : array
            (M, 3,) array of positions.
        time : array or float
            (K,) array of times (in seconds).
        alpha : float
            Ground thermal diffusivity (in m^2/s).
        r_min : float, default: ``0.``
            Minimum distance (in meters) added to the distance between the
            heat source and the positions `p`.

        Returns
        -------
        array
            (K, M, N,) array of the point heat source solution. If `time`
            is a ``float``, then ``K=0``.
        """
        # Coordinates (xi) of all sources at local segment coordinates (xi')
        xi = self.f_xi_sb(xi_p)
        # Point heat source solutions
        h = self.path.point_heat_source(xi, p, time, alpha, r_min=r_min) * self.segment_ratios
        return h

    @classmethod
    def from_dimensions(cls, L: float, D: float, r_b: float, x: float, y: float, basis: Basis, n_segments: int, tilt: float = 0., orientation: float = 0., segment_ratios: ArrayLike | None = None) -> Self:
        """Straight borehole from its dimensions.

        Parameters
        ----------
        L : float
            Borehole length (in meters).
        D : float
            Borehole buried depth (in meters).
        r_b : float
            Borehole radius (in meters).
        x, y : float
            Horizontal position (in meters) of the top end of the
            borehole.
        basis : basis
            Basis functions.
        n_segments : int
            Number of segments.
        tilt : float, default: ``0.``
            Tilt angle (in radians) of the borehole with respect to
            vertical.
        orientation : float, default: ``0.``
            Orientation (in radians) of the inclined borehole. An
            inclination toward the x-axis corresponds to an orientation
            of zero.
        segment_ratios : array_like or None, default: None
            Normalized size of the segments. Should total ``1``
            (i.e. ``sum(segment_ratios) = 1``). If `segment_ratios` is
            ``None``, segments of equal size are considered (i.e.
            ``segment_ratios[v] = 1 / n_segments``).

        Returns
        -------
        borehole
            Instance of the `Borehole` class.

        """
        xi = jnp.array([-1., 1.])
        p = jnp.array(
            [
                [x, y, -D],
                [x + L * jnp.sin(tilt) * jnp.cos(orientation), y + L * jnp.sin(tilt) * jnp.sin(orientation), -D - L * jnp.cos(tilt)],
            ]
        )
        path = Path(xi, p)
        return cls(r_b, path, basis, n_segments, segment_ratios=segment_ratios)



# -*- coding: utf-8 -*-
from typing import Tuple

from jax import numpy as jnp
from matplotlib import pyplot as plt
from matplotlib.axes import Axes
from matplotlib.figure import Figure

from ._format_axis import _format_axis
from ..ground_heat_exchanger import (
CoaxialHeatExchanger,
UTubeHeatExchanger
)


def plot_ground_heat_exchanger(ground_heat_exchanger: CoaxialHeatExchanger | CoaxialHeatExchanger, ax: Axes | None = None) -> Axes | Tuple[Figure, Axes]:
    """Plot a ground heat exchanger.

    Parameters
    ----------
    ground_heat_exchanger : ground heat exchanger
        Ground heat exchanger to be plotted.
    ax : axes, None, default: ``None``
        The axis on which to draw the ground heat exchanger. If `ax` is
        ``None``, a new figure and new axes are created.

    Returns
    -------
    fig : figure
        The figure. Only return if `ax` is ``None``.
    ax : axes
        The axis.

    """
    ax_is_none = ax is None
    if ax is None:
        fig, ax = plt.subplots(layout='constrained')
    _format_axis(ax, axis_labels=[r'$x$ [m]', r'$y$ [m]'], equal=True)

    if isinstance(ground_heat_exchanger, CoaxialHeatExchanger):
        _plot_coaxial_heat_exchanger(ground_heat_exchanger, ax=ax)
    elif isinstance(ground_heat_exchanger, UTubeHeatExchanger):
        _plot_u_tube_heat_exchanger(ground_heat_exchanger, ax=ax)

    if ax_is_none:
        return fig, ax
    else:
        return ax


def _plot_coaxial_heat_exchanger(coaxial_heat_exchanger: CoaxialHeatExchanger, ax: Axes | None = None) -> Axes | Tuple[Figure, Axes]:
    """Plot a U-tube ground heat exchanger.

    Parameters
    ----------
    coaxial_heat_exchanger : coaxial ground heat exchanger
        Ground heat exchanger to be plotted.
    ax : axes, None, default: ``None``
        The axis on which to draw the ground heat exchanger. If `ax` is
        ``None``, a new figure and new axes are created.

    Returns
    -------
    fig : figure
        The figure. Only return if `ax` is ``None``.
    ax : axes
        The axis.

    """
    ax_is_none = ax is None
    if ax is None:
        fig, ax = plt.subplots(layout='constrained')
    _format_axis(ax, axis_labels=[r'$x$ [m]', r'$y$ [m]'], equal=True)

    ghe = coaxial_heat_exchanger

    # Color cycle
    prop_cycle = plt.rcParams['axes.prop_cycle']
    colors = prop_cycle.by_key()['color']

    # Draw dots at min/max x/y for matplotlib to size the figure correctly
    ax.plot(
        [-ghe.r_b, 0., ghe.r_b, 0.],
        [0., ghe.r_b, 0., -ghe.r_b],
        'k.',
        alpha=0.
    )
    # Borehole wall outline
    borehole_wall = plt.Circle(
        (0., 0.),
        radius=ghe.r_b,
        fill=False,
        color='k',
        linestyle='--'
        )
    # Add borehole wall patch
    ax.add_patch(borehole_wall)

    # Pipes
    n_pipes_over_two = ghe._n_pipes_over_two
    for i in range(n_pipes_over_two):
        # Coordinates of pipes
        (x, y) = ghe.p[i]
        # Identify inlet
        if i in ghe.indices_inner:
            dx_inlet = 0.
            dx_outlet = 0.5 * ghe.r_p_out[i] + 0.5 * ghe.r_p_in[i + n_pipes_over_two]
        else:
            dx_inlet = 0.5 * ghe.r_p_in[i] + 0.5 * ghe.r_p_out[i + n_pipes_over_two]
            dx_outlet = 0.

        # Pipe outline (inlet)
        pipe_in_in = plt.Circle(
            (x, y),
            radius=ghe.r_p_in[i],
            fill=False,
            linestyle='-',
            color=colors[i]
            )
        pipe_in_out = plt.Circle(
            (x, y),
            radius=ghe.r_p_out[i],
            fill=False,
            linestyle='-',
            color=colors[i]
            )
        ax.text(
            x + dx_inlet,
            y,
            i,
            ha="center",
            va="center")

        # Pipe outline (outlet)
        pipe_out_in = plt.Circle(
            (x, y),
            radius=ghe.r_p_in[i + n_pipes_over_two],
            fill=False,
            linestyle='-',
            color=colors[i]
            )
        pipe_out_out = plt.Circle(
            (x, y),
            radius=ghe.r_p_out[i + n_pipes_over_two],
            fill=False,
            linestyle='-',
            color=colors[i]
            )
        ax.text(
            x + dx_outlet,
            y,
            i + n_pipes_over_two,
            ha="center",
            va="center")

        # Add pipe patches
        ax.add_patch(pipe_in_in)
        ax.add_patch(pipe_in_out)
        ax.add_patch(pipe_out_in)
        ax.add_patch(pipe_out_out)

    if ax_is_none:
        return fig, ax
    else:
        return ax


def _plot_u_tube_heat_exchanger(u_tube_heat_exchanger: UTubeHeatExchanger, ax: Axes | None = None) -> Axes | Tuple[Figure, Axes]:
    """Plot a U-tube ground heat exchanger.

    Parameters
    ----------
    u_tube_heat_exchanger : U-tube ground heat exchanger
        Ground heat exchanger to be plotted.
    ax : axes, None, default: ``None``
        The axis on which to draw the ground heat exchanger. If `ax` is
        ``None``, a new figure and new axes are created.

    Returns
    -------
    fig : figure
        The figure. Only return if `ax` is ``None``.
    ax : axes
        The axis.

    """
    ax_is_none = ax is None
    if ax is None:
        fig, ax = plt.subplots(layout='constrained')
    _format_axis(ax, axis_labels=[r'$x$ [m]', r'$y$ [m]'], equal=True)

    ghe = u_tube_heat_exchanger

    # Color cycle
    prop_cycle = plt.rcParams['axes.prop_cycle']
    colors = prop_cycle.by_key()['color']

    # Draw dots at min/max x/y for matplotlib to size the figure correctly
    ax.plot(
        [-ghe.r_b, 0., ghe.r_b, 0.],
        [0., ghe.r_b, 0., -ghe.r_b],
        'k.',
        alpha=0.
    )
    # Borehole wall outline
    borehole_wall = plt.Circle(
        (0., 0.),
        radius=ghe.r_b,
        fill=False,
        color='k',
        linestyle='--'
        )
    # Add borehole wall patch
    ax.add_patch(borehole_wall)

    # Pipes
    n_pipes_over_two = ghe._n_pipes_over_two
    for i in range(n_pipes_over_two):
        # Coordinates of pipes
        (x_in, y_in) = ghe.p[i]
        (x_out, y_out) = ghe.p[i + n_pipes_over_two]

        # Pipe outline (inlet)
        pipe_in_in = plt.Circle(
            (x_in, y_in),
            radius=ghe.r_p_in[i],
            fill=False,
            linestyle='-',
            color=colors[i]
            )
        pipe_in_out = plt.Circle(
            (x_in, y_in),
            radius=ghe.r_p_out[i],
            fill=False,
            linestyle='-',
            color=colors[i]
            )
        ax.text(
            x_in,
            y_in,
            i,
            ha="center",
            va="center")

        # Pipe outline (outlet)
        pipe_out_in = plt.Circle(
            (x_out, y_out),
            radius=ghe.r_p_in[i + n_pipes_over_two],
            fill=False,
            linestyle='-',
            color=colors[i]
            )
        pipe_out_out = plt.Circle(
            (x_out, y_out),
            radius=ghe.r_p_out[i + n_pipes_over_two],
            fill=False,
            linestyle='-',
            color=colors[i]
            )
        ax.text(
            x_out,
            y_out,
            i + n_pipes_over_two,
            ha="center",
            va="center")

        # Add pipe patches
        ax.add_patch(pipe_in_in)
        ax.add_patch(pipe_in_out)
        ax.add_patch(pipe_out_in)
        ax.add_patch(pipe_out_out)

    if ax_is_none:
        return fig, ax
    else:
        return ax



# -*- coding: utf-8 -*-
from abc import ABC, abstractmethod
from functools import partial

from jax import Array, jit, vmap


class _GroundHeatExchanger(ABC):
    """Ground heat exchanger cross-section.

    Attributes
    ----------
    n_pipes : int
        Number of pipes.

    """

    @abstractmethod
    def thermal_resistances(self, m_flow: float) -> Array:
        """Evaluate delta-circuit thermal resistances.

        Parameters
        ----------
        m_flow : float
            Fluid mass flow rate (in kg/s).

        Returns
        -------
        array
            (`n_pipes`, `n_pipes`,) array of delta-circuit thermal
            resistances (in m-K/W).

        """
        ...



from ._format_axis import _format_axis
from .plot_basis import plot_basis
from .plot_borehole import plot_borehole
from .plot_borefield import plot_borefield
from .plot_gfunction import plot_gfunction
from .plot_ground_heat_exchanger import plot_ground_heat_exchanger
from .plot_path import plot_path



# -*- coding: utf-8 -*-
from typing import List, Self

from jax import numpy as jnp
from jax import Array, jit, vmap
from jax.typing import ArrayLike

from ..basis import Basis
from ..borehole import Borehole
from ..path import Path


class Borefield:
    """Borefield.

    Parameters
    ----------
    boreholes : list of borehole
        Boreholes in the borefield.

    Attributes
    ----------
    n_boreholes : int
        Number of boreholes.
    n_nodes : int
        Number of nodes per borehole.
    L : array
        Borehole lengths (in meters).
    xi : array
        Coordinates of the nodes. They are the same for all boreholes.
    p : array
        (`n_boreholes`, `n_nodes`, 3,) array of node positions.
    dp_dxi : array
        (`n_boreholes`, `n_nodes`, 3,) array of the derivatives of the
        position at the node coordinates.
    J : array
        (`n_boreholes`, `n_nodes`,) array of the norm of the Jacobian at
        the node coordinates.
    s : array
        (`n_boreholes`, `n_nodes`,) array of the longitudinal position at
        the node coordinates.
    w : array
        (`n_boreholes`, `n_nodes`,) array of quadrature weights at the
        node coordinates. These quadrature weights take into account the
        norm of the Jacobian.

    """

    def __init__(self, boreholes: List[Borehole]):
        self.boreholes = boreholes
        self.n_boreholes = len(boreholes)
        self.n_nodes = boreholes[0].n_nodes
        self.L = jnp.array([borehole.L for borehole in self.boreholes])

        # --- Basis functions ---
        self.f_psi = jit(
            lambda _eta: jnp.stack(
                [
                    vmap(borehole.f_psi, in_axes=0)(_eta) if len(jnp.shape(_eta)) > 0 else borehole.f_psi(_eta)
                    for borehole in boreholes
                ],
                axis=-2
            )
        )
        self.f = jit(
            lambda _eta, f_nodes: vmap(jnp.dot, in_axes=(-2, 0), out_axes=0)(self.f_psi(_eta), f_nodes)
        )

        # --- Nodal values of path and basis functions ---
        # Borehole coordinates (xi)
        self.xi = self.boreholes[0].xi
        # Positions (p)
        self.p = jnp.stack([borehole.p for borehole in self.boreholes], axis=0)
        # Derivatives of position (dp/dxi)
        self.dp_dxi = jnp.stack([borehole.dp_dxi for borehole in self.boreholes], axis=0)
        # Norms of the Jacobian (J)
        self.J = jnp.stack([borehole.J for borehole in self.boreholes], axis=0)
        # Longitudinal positions (s)
        self.s = jnp.stack([borehole.s for borehole in self.boreholes], axis=0)
        # Integration weights
        self.w = jnp.stack([borehole.w for borehole in self.boreholes], axis=0)

    def h_to_self(self, time: Array, alpha: float) -> Array:
        """Thermal response factors to its own nodes.

        Parameters
        ----------
        time : array
            (K,) array of times (in seconds).
        alpha : float
            Ground thermal diffusivity (in m^2/s).

        Returns
        -------
        array
            (K, `n_boreholes`, `n_nodes`, `n_boreholes`, `n_nodes`,) array
            of thermal response factors.
        """
        n_boreholes = self.n_boreholes
        h_to_self = jnp.stack(
            [
                jnp.stack(
                    [
                        self.boreholes[j].h_to_self(time, alpha) if i == j else self.boreholes[j].h_to_borehole(self.boreholes[i], time, alpha)
                        for j in jnp.arange(n_boreholes)
                    ],
                    axis=-2)
                for i in jnp.arange(n_boreholes)
            ],
            axis=1)
        return h_to_self

    def h_to_point(self, p: Array, time: Array, alpha: float, r_min: float = 0.) -> Array:
        """Thermal response factors to a point.

        Parameters
        ----------
        p : array
            (M, 3,) array of the positions at which thermal response
            factors are evaluated.
        time : array
            (K,) array of times (in seconds).
        alpha : float
            Ground thermal diffusivity (in m^2/s).
        r_min : float, default: ``0.``
            Minimum distance (in meters) added to the distance between the
            heat source and the positions `p`.

        Returns
        -------
        array
            (K, M, `n_boreholes`, `n_nodes`,) array of thermal response
            factors.
        """
        n_boreholes = self.n_boreholes
        h_to_point = jnp.stack(
            [
                self.boreholes[j].h_to_point(p, time, alpha, r_min=r_min) for j in jnp.arange(n_boreholes)
                ],
            axis=-2)
        return h_to_point

    @classmethod
    def from_positions(cls, L: ArrayLike, D: ArrayLike, r_b: ArrayLike, x: ArrayLike, y: ArrayLike, basis: Basis, n_segments: int, tilt: ArrayLike = 0., orientation: ArrayLike = 0., segment_ratios: ArrayLike | None = None) -> Self:
        """Field of straight boreholes from their dimensions.

        Parameters
        ----------
        L : array_like
            Borehole length (in meters).
        D : array_like
            Borehole buried depth (in meters).
        r_b : array_like
            Borehole radius (in meters).
        x, y : array_like
            Horizontal position (in meters) of the top end of the
            borehole.
        basis : basis
            Basis functions.
        n_segments : int
            Number of segments per borehole.
        tilt : array_like, default: ``0.``
            Tilt angle (in radians) of the boreholes with respect to
            vertical.
        orientation : array_like, default: ``0.``
            Orientation (in radians) of the inclined boreholes. An
            inclination toward the x-axis corresponds to an orientation
            of zero.
        segment_ratios : array_like or None, default: None
            Normalized size of the segments. Should total ``1``
            (i.e. ``sum(segment_ratios) = 1``). If `segment_ratios` is
            ``None``, segments of equal size are considered (i.e.
            ``segment_ratios[v] = 1 / n_segments``).

        Returns
        -------
        borefield
            Instance of the `Borefield` class.

        """
        # Runtime type validation
        if not isinstance(x, ArrayLike):
            raise TypeError(f"Expected arraylike input; got {x}")
        if not isinstance(y, ArrayLike):
            raise TypeError(f"Expected arraylike input; got {y}")
        # Convert input to jax.Array
        x = jnp.asarray(x)
        y = jnp.asarray(y)

        n_boreholes = len(x)
        L = jnp.broadcast_to(L, n_boreholes)
        D = jnp.broadcast_to(D, n_boreholes)
        r_b = jnp.broadcast_to(r_b, n_boreholes)
        tilt = jnp.broadcast_to(tilt, n_boreholes)
        orientation = jnp.broadcast_to(orientation, n_boreholes)
        boreholes = []
        xi = jnp.array([-1., 1.])
        for j in range(n_boreholes):
            p = jnp.array(
                [
                    [x[j], y[j], -D[j]],
                    [x[j] + L[j] * jnp.sin(tilt[j]) * jnp.cos(orientation[j]), y[j] + L[j] * jnp.sin(tilt[j]) * jnp.sin(orientation[j]), -D[j] - L[j] * jnp.cos(tilt[j])],
                ]
            )
            path = Path(xi, p, order=order)
            boreholes.append(Borehole(r_b[j], path, basis, n_segments, segment_ratios=segment_ratios))
        return cls(boreholes)

    @classmethod
    def rectangle_field(cls, N_1: int, N_2: int, B_1: float, B_2: float, L: float, D: float, r_b: float, basis: Basis, n_segments: int, segment_ratios: ArrayLike | None = None, order: int | None = None) -> Self:
        """Field of vertical boreholes in a rectangular configuration.

        Parameters
        ----------
        N_1, N_2 : int
            Number of columns and rows in the borefield.
        B_1, B_2 : float
            Spacing between columns and rows (in meters).
        L : float
            Borehole length (in meters).
        D : float
            Borehole buried depth (in meters).
        r_b : float
            Borehole radius (in meters).
        basis : basis
            Basis functions.
        n_segments : int
            Number of segments per borehole.
        segment_ratios : array_like or None, default: None
            Normalized size of the segments. Should total ``1``
            (i.e. ``sum(segment_ratios) = 1``). If `segment_ratios` is
            ``None``, segments of equal size are considered (i.e.
            ``segment_ratios[v] = 1 / n_segments``).

        Returns
        -------
        borefield
            Instance of the `Borefield` class.

        """
        # Borehole positions and orientation
        x = jnp.tile(jnp.arange(N_1), N_2) * B_1
        y = jnp.repeat(jnp.arange(N_2), N_1) * B_2
        return cls.from_positions(L, D, r_b, x, y, basis, n_segments, segment_ratios=segment_ratios)
        



# -*- coding: utf-8 -*-
from functools import partial

from jax import numpy as jnp
from jax import Array, jit, vmap
from jax.lax import switch

from .darcy_friction_factor import darcy_friction_factor
from .reynolds_number import (
reynolds_number_annular_pipe,
reynolds_number_circular_pipe
)


def convective_heat_transfer_coefficient_annular_pipe(m_flow: float, r_p_in: float, r_p_out: float, mu_f: float, rho_f: float, k_f: float, cp_f: float, epsilon: float) -> Array:
    """Convective heat transfer coefficient in circular pipe.

    Parameters
    ----------
    m_flow : float
        Fluid mass flow rate (in kg/s).
    r_p_in : float
        Outer radius of the inner pipe (in meters).
    r_p_out : float
        Inner radius of the outer pipe (in meters).
    mu_f : float
        Fluid dynamic viscosity (in kg/m-s).
    rho_f : float
        Fluid density (in kg/m3).
    k_f : float
        Fluid thermal conductivity (in W/m-K).
    cp_f : float
        Fluid isobaric specific heat capacity (in J/kg-K).
    epsilon : float
        Pipe surface roughness (in meters).

    Returns
    -------
    array
        Convective heat transfer coefficients at the inner and outer
        surfaces of the annulus (in W/m2-K).
    
    """
    # Reynolds number
    Re = reynolds_number_annular_pipe(m_flow, r_p_in, r_p_out, mu_f, rho_f)
    # Prandtl number
    Pr = cp_f * mu_f / k_f
    # Darcy friction factor
    D = 2 * (r_p_out - r_p_in)
    E = epsilon / D
    f_darcy = darcy_friction_factor(Re, E)
    # Nusselt number
    r_ratio = r_p_in / r_p_out
    Nu = nusselt_number_annular_pipe(Re, Pr, f_darcy, r_ratio)
    # Convective heat transfer coefficient
    h_fluid = k_f * Nu / D
    return h_fluid


def convective_heat_transfer_coefficient_circular_pipe(m_flow: float, r_p: float, mu_f: float, rho_f: float, k_f: float, cp_f: float, epsilon: float) -> float:
    """Convective heat transfer coefficient in circular pipe.

    Parameters
    ----------
    m_flow : float
        Fluid mass flow rate (in kg/s).
    r_p : float
        Inner radius of the pipe (in meters).
    mu_f : float
        Fluid dynamic viscosity (in kg/m-s).
    rho_f : float
        Fluid density (in kg/m3).
    k_f : float
        Fluid thermal conductivity (in W/m-K).
    cp_f : float
        Fluid isobaric specific heat capacity (in J/kg-K).
    epsilon : float
        Pipe surface roughness (in meters).

    Returns
    -------
    float
        Convective heat transfer coefficient (in W/m2-K).
    
    """
    # Reynolds number
    Re = reynolds_number_circular_pipe(m_flow, r_p, mu_f, rho_f)
    # Prandtl number
    Pr = cp_f * mu_f / k_f
    # Darcy friction factor
    D = 2 * r_p
    E = epsilon / D
    f_darcy = darcy_friction_factor(Re, E)
    # Nusselt number
    Nu = nusselt_number_circular_pipe(Re, Pr, f_darcy)
    # Convective heat transfer coefficient
    h_fluid = k_f * Nu / D
    return h_fluid


def nusselt_number_annular_pipe(Re: float, Pr: float, f_darcy: float, r_ratio: float) -> Array:
    """Nusselt numbers in annular pipe.

    Parameters
    ----------
    Re : float
        Reynolds number.
    Pr : float
        Prandtl number.
    f_darcy : float
        Darcy friction factor.
    r_ratio : float
        Ratio of inner and outer radii or the annulus.

    Returns
    -------
    array
        Nusselt number at the inner and outer pipe surfaces.

    """
    Re_crit_lower = 2300.
    Re_crit_upper = 4000.
    # Nusselt number for laminar flow
    def _nusselt_number_laminar_flow(_Re: float) -> Array:
        Nu_laminar = 3.66 + 1.2 * r_ratio**jnp.array([-0.8, 0.5])
        return Nu_laminar
    # Nusselt number for turbulent flow
    def _nusselt_number_turbulent_flow(_Re: float) -> Array:
        Nu_turbulent = nusselt_number_turbulent_flow(_Re, Pr, f_darcy)
        return jnp.broadcast_to(Nu_turbulent, 2)
    # Nusselt number in transition regime
    Nu_crit_lower = _nusselt_number_laminar_flow(Re_crit_lower)
    Nu_crit_upper = _nusselt_number_turbulent_flow(Re_crit_upper)
    def _nusselt_number_transition(_Re: float) -> Array:
        Nu_transition = nusselt_number_transition(
            _Re, Re_crit_lower, Re_crit_upper, Nu_crit_lower, Nu_crit_upper)
        return Nu_transition
    # Piecewise evaluation of Nusselt number
    funclist = [
        _nusselt_number_laminar_flow,
        _nusselt_number_transition,
        _nusselt_number_turbulent_flow
    ]
    bounds = jnp.array([Re_crit_lower, Re_crit_upper])
    index = jnp.searchsorted(bounds, Re)
    Nu = switch(index, funclist, Re)
    return Nu


def nusselt_number_circular_pipe(Re: float, Pr: float, f_darcy: float) -> float:
    """Nusselt number in circular pipe.

    Parameters
    ----------
    Re : float
        Reynolds number.
    Pr : float
        Prandtl number.
    f_darcy : float
        Darcy friction factor.

    Returns
    -------
    float
        Nusselt number.

    """
    Re_crit_lower = 2300.
    Re_crit_upper = 4000.
    # Condition on the flow regime
    condlist = [Re <= Re_crit_lower, Re >= Re_crit_upper]
    # Nusselt number for laminar flow
    Nu_laminar = 3.66
    # Nusselt number for turbulent flow
    _nusselt_number_turbulent_flow = partial(
        nusselt_number_turbulent_flow,
        Pr=Pr,
        f_darcy=f_darcy)
    # Nusselt number at the onset of turbulent flow
    Nu_crit_upper = nusselt_number_turbulent_flow(
        Re_crit_upper, Pr, f_darcy)
    # Nusselt number in transition regime
    _nusselt_number_transition = partial(
        nusselt_number_transition,
        Re_crit_lower=Re_crit_lower,
        Re_crit_upper=Re_crit_upper,
        Nu_crit_lower=Nu_laminar,
        Nu_crit_upper=Nu_crit_upper)
    # Piecewise evaluation of Nusselt number
    funclist = [Nu_laminar, _nusselt_number_turbulent_flow, _nusselt_number_transition]
    Nu = jnp.piecewise(jnp.float32(Re), condlist, funclist)
    return Nu


def nusselt_number_transition(Re: float, Re_crit_lower: float, Re_crit_upper: float, Nu_crit_lower: float | Array, Nu_crit_upper: float | Array) -> float | Array:
    """Nusselt number for turbulent flow in any pipe.

    Parameters
    ----------
    Re : float
        Reynolds number.
    Re_crit_lower : float
        Reynolds number at the lower limit of the transition regime.
    Re_crit_upper : float
        Reynolds number at the upper limit of the transition regime.
    Nu_crit_lower : float
        Nusselt number at the lower limit of the transition regime.
    Nu_crit_upper : float
        Nusselt number at the upper limit of the transition regime.

    Returns
    -------
    float
        Nusselt number.

    """
    gamma = (Re - Re_crit_lower) / (Re_crit_upper - Re_crit_lower)
    Nu = (1 - gamma) * Nu_crit_lower + gamma * Nu_crit_upper
    return Nu


def nusselt_number_turbulent_flow(Re: float, Pr: float, f_darcy: float) -> float:
    """Nusselt number for turbulent flow in any pipe.

    Parameters
    ----------
    Re : float
        Reynolds number.
    Pr : float
        Prandtl number.
    f_darcy : float
        Darcy friction factor.

    Returns
    -------
    float
        Nusselt number.

    """
    Nu = 0.125 * f_darcy * (Re - 1.0e3) * Pr / (
        1 + 12.7 * jnp.sqrt(0.125 * f_darcy) * (Pr**(2 / 3) - 1)
    )
    return Nu



# -*- coding: utf-8 -*-
from abc import ABC, abstractmethod

from jax import numpy as jnp
from jax import Array, jit


class _TemporalSuperposition(ABC):
    """Temporal superposition of thermal response factors.

    """

    @abstractmethod
    def next_time_step(self) -> float:
        """Advance to next simulation time step.

        Returns
        -------
        float
            Time (in seconds) of the new time step.

        """
        ...

    @abstractmethod
    def reset_history(self):
        """Reset the history to its initial condition.

        """
        ...

    @abstractmethod
    def set_current_load(self, q: Array):
        """Set the current heat extraction rate.

        Parameters
        ----------
        q : array
            (`n_boreholes`, `n_nodes`,) array of heat extraction rates
            (in W/m).

        """
        ...

    @abstractmethod
    def temperature(self) -> Array:
        """Evaluate temperatures at nodes.

        Returns
        ----------
        T_b : array
            (`n_boreholes`, `n_nodes`,) array of borehole wall
            temperature variations (in degree Celcius).

        """
        ...

    @abstractmethod
    def temperature_to_point(self) -> Array:
        """Evaluate temperatures at point.

        Returns
        ----------
        T : array
            (`n_points`,) array of ground temperature variations
            (in degree Celcius).

        """
        ...

    @staticmethod
    @jit
    def _temperature(h_to_self: Array, q: Array) -> Array:
        """Spatial and temporal superposition at nodes.

        Parameters
        ----------
        h_to_self : array
            (`n_times`, `n_boreholes`, `n_nodes`, `n_boreholes`,
            `n_nodes`,) array of thermal response factors.
        q : array
            (`n_times`, `n_boreholes`, `n_nodes`,) array of heat
            extraction rates (in W/m).

        Returns
        ----------
        T_b : array
            (`n_boreholes`, `n_nodes`,) array of borehole wall
            temperatures (in degree Celcius).

        """
        T = jnp.tensordot(h_to_self, q, axes=([0, -2, -1], [0, -2, -1]))
        return T

    @staticmethod
    @jit
    def _temperature_to_point(h_to_point: Array, q: Array) -> Array:
        """Spatial and temporal superposition at points.

        Parameters
        ----------
        h_to_point : array
            (`n_times`, `n_points`, `n_boreholes`, `n_nodes`,) array of
            thermal response factors.
        q : array
            (`n_times`, `n_boreholes`, `n_nodes`,) array of heat
            extraction rates (in W/m).

        Returns
        ----------
        T : array
            (`n_points`,) array of ground temperatures (in degree
            Celcius).

        """
        T = jnp.tensordot(h_to_point, q, axes=([0, -2, -1], [0, -2, -1]))
        return T



# -*- coding: utf-8 -*-
from collections.abc import Callable
from functools import partial
from typing import Self, Tuple

from jax import numpy as jnp
from jax import Array, jit, vmap
from jax.scipy.linalg import expm
from jax.typing import ArrayLike

from ._tube import _Tube
from ..basis import Basis
from ..path import Path


class MultipleUTube(_Tube):
    """Multiple U-Tube geothermal borehole.

    Parameters
    ----------
    R_d : array_like or callable
        (`n_pipes`, `n_pipes`,) array of thermal resistances (in m-K/W),
        or callable that takes the mass flow rate as input (in kg/s) and
        returns a (`n_pipes`, `n_pipes`,) array.
    r_b : float
        Borehole radius (in meters).
    path : path
        Path of the borehole.
    basis : basis
        Basis functions.
    n_segments : int
        Number of segments.
    segment_ratios : array_like or None, default: None
        Normalized size of the segments. Should total ``1``
        (i.e. ``sum(segment_ratios) = 1``). If `segment_ratios` is
        ``None``, segments of equal size are considered (i.e.
        ``segment_ratios[v] = 1 / n_segments``).

    Attributes
    ----------
    n_pipes : int
        Number of pipes in the borehole.
    n_nodes : int
        Total number of nodes along the borehole.
    xi_edges : array
        (`n_segments`+1,) array of the coordinates of the edges of the
        segments.
    L : float
        Borehole length (in meters).
    xi : array
        (`n_nodes`,) array of node coordinates.
    p : array
        (`n_nodes`, 3,) array of node positions.
    dp_dxi : array
        (`n_nodes`, 3,) array of the derivatives of the position at the
        node coordinates.
    J : array
        (`n_nodes`,) array of the norm of the Jacobian at the node
        coordinates.
    s : array
        (`n_nodes`,) array of the longitudinal position at the node
        coordinates.
    w : array
        (`n_nodes`,) array of quadrature weights at the node coordinates.
        These quadrature weights take into account the norm of the
        Jacobian.

    """

    def _fluid_temperature_a_in(self, xi: Array | float, beta_ij: Array) -> Array:
        """Inlet coefficient to evaluate the fluid temperatures.

        Parameters
        ----------
        xi : array or float
            (M,) array of coordinates along the borehole.
        beta_ij: array
            (`n_pipes`, `n_pipes`,) array of thermal conductance
            coefficients.

        Returns
        -------
        array
            (M, `n_pipes`,) array of coefficients for the inlet fluid
            temperature.

        """
        # General solution at coordinates `xi`
        E_0_xi = self._general_solution_a_0(xi, beta_ij)
        # General solution at the bottom-end
        E_0 = self._general_solution_a_0(1., beta_ij)
        # Coefficients for connectivity at the top of the borehole
        # T_f_d(-1) = c_in * T_f_in + c_u @ T_f_u(-1)
        c_in = self._top_connectivity_in
        c_u = self._top_connectivity_u

        # Intermediate coefficients
        delta_E_0_xi = (
            E_0_xi[..., :self._n_pipes_over_two] @ c_u
            + E_0_xi[..., self._n_pipes_over_two:]
        )
        delta_E_0_u = (
            E_0[:self._n_pipes_over_two]
            - E_0[self._n_pipes_over_two:]
        )
        delta_E_0_u = (
            delta_E_0_u[:, :self._n_pipes_over_two] @ c_u
            + delta_E_0_u[:, self._n_pipes_over_two:]
        )
        delta_E_0_in = (
            E_0[self._n_pipes_over_two:]
            - E_0[:self._n_pipes_over_two]
        )
        delta_E_0_in = delta_E_0_in[:, :self._n_pipes_over_two] @ c_in

        # Inlet coefficient to evaluate the fluid temperatures
        a_in = (
            delta_E_0_xi @ jnp.linalg.solve(delta_E_0_u, delta_E_0_in)
            + E_0_xi[..., :self._n_pipes_over_two] @ c_in
        )
        return a_in

    def _fluid_temperature_a_b(self, xi: Array | float, beta_ij: Array) -> Array:
        """Borehole wall coefficient to evaluate the fluid temperatures.

        Parameters
        ----------
        xi : array or float
            (M,) array of coordinates along the borehole.
        beta_ij: array
            (`n_pipes`, `n_pipes`,) array of thermal conductance
            coefficients.

        Returns
        -------
        array
            (M, `n_nodes`,) array of coefficients for the borehole wall
            temperature.

        """
        # General solution at coordinates `xi`
        E_0_xi = self._general_solution_a_0(xi, beta_ij)
        E_b_xi = self._general_solution_a_b(xi, beta_ij)
        # General solution at the bottom-end
        E_0 = self._general_solution_a_0(1., beta_ij)
        E_b = self._general_solution_a_b(1., beta_ij)
        # Coefficients for connectivity at the top of the borehole
        # T_f_d(-1) = c_in * T_f_in + c_u @ T_f_u(-1)
        c_in = self._top_connectivity_in
        c_u = self._top_connectivity_u

        # Intermediate coefficients
        delta_E_0_xi = (
            E_0_xi[..., :self._n_pipes_over_two] @ c_u
            + E_0_xi[..., self._n_pipes_over_two:]
        )
        delta_E_0_u = (
            E_0[:self._n_pipes_over_two]
            - E_0[self._n_pipes_over_two:]
        )
        delta_E_0_u = (
            delta_E_0_u[:, :self._n_pipes_over_two] @ c_u
            + delta_E_0_u[:, self._n_pipes_over_two:]
        )
        delta_E_b = (
            E_b[self._n_pipes_over_two:]
            - E_b[:self._n_pipes_over_two]
        )

        # Borehole wall coefficient to evaluate the fluid temperatures
        a_b = (
            delta_E_0_xi @ jnp.linalg.solve(delta_E_0_u, delta_E_b)
            + E_b_xi
        )
        return a_b

    def _general_solution(self, xi: Array | float, m_flow: float, cp_f: float) -> Tuple[Array, Array]:
        """Coefficients to evaluate the general solution.

        Parameters
        ----------
        xi : array or float
            (M,) array of coordinates along the borehole.
        m_flow : float
            Fluid mass flow rate (in kg/s).
        cp_f : float
            Fluid specific isobaric heat capacity (in J/kg-K).

        Returns
        -------
        a_0 : array or float
            (M, `n_pipes`,) array of coefficients for the top-end fluid
            temperature.
        a_b : array
            (M, `n_pipes`, `n_nodes`,) array of coefficients for the
            borehole wall temperature.

        """
        beta_ij = self._beta_ij(m_flow, cp_f)
        a_0 = self._general_solution_a_0(xi, beta_ij)
        a_b = self._general_solution_a_b(xi, beta_ij)
        return a_0, a_b

    def _general_solution_a_0(self, xi: Array | float, beta_ij: Array) -> Array:
        """Top-end coefficient to evaluate the general solution.

        Parameters
        ----------
        xi : array or float
            (M,) array of coordinates along the borehole.
        beta_ij: array
            (`n_pipes`, `n_pipes`,) array of thermal conductance
            coefficients.

        Returns
        -------
        array or float
            (M, `n_pipes`, `n_pipes`) array of coefficients for the
            top-end fluid temperature.

        """
        if len(jnp.shape(xi)) == 1:
            a_0 = vmap(
                self._general_solution_a_0,
                in_axes=(0, None),
                out_axes=0
            )(xi, beta_ij)
        else:
            s = self.path.f_s(xi)
            a_0 = expm(self._ode_coefficients(beta_ij) * s)
        return a_0

    def _general_solution_a_b(self, xi: Array | float, beta_ij: Array) -> Array:
        """Borehole wall coefficient to evaluate the general solution.

        Parameters
        ----------
        xi : array or float
            (M,) array of coordinates along the borehole.
        beta_ij: array
            (`n_pipes`, `n_pipes`,) array of thermal conductance
            coefficients.

        Returns
        -------
        array
            (M, `n_pipes`, `n_nodes`,) array of coefficients for the
            borehole wall temperature.

        """
        if len(jnp.shape(xi)) == 1:
            a_b = vmap(
                self._general_solution_a_b,
                in_axes=(0, None),
                out_axes=0
            )(xi, beta_ij)
        else:
            s = self.path.f_s(xi)
            high = jnp.maximum(-1., jnp.minimum(1., self.f_xi_bs(xi)))
            a, b = self.xi_edges[:-1], self.xi_edges[1:]
            f_xi_bs = lambda _eta, _a, _b: 0.5 * (_b + _a) + 0.5 * _eta * (_b - _a)

            ode_coefficients = self._ode_coefficients(beta_ij)
            ode_coefficients_sum = ode_coefficients.sum(axis=1)
            expm_delta_s = lambda _eta: expm(ode_coefficients * (s - self.path.f_s(_eta)))
            integrand = lambda _eta: -expm_delta_s(_eta) @ ode_coefficients_sum * self.path.f_J(_eta)
            integrand_segment = lambda _eta, _a, _b, _ratio: (
                integrand(f_xi_bs(_eta, _a, _b)) * _ratio
            )
            integral = lambda _a, _b, _ratio, _high: self.basis.quad_gl(
                    vmap(
                        lambda _eta: integrand_segment(_eta, _a, _b, _ratio),
                        in_axes=0,
                        out_axes=-1
                    ),
                -1.,
                _high
                )
            a_b = vmap(
                    integral,
                    in_axes=(0, 0, 0, 0),
                    out_axes=1
                )(a, b, self.segment_ratios, high).reshape(self.n_pipes, self.n_nodes)
        return a_b

    def _ode_coefficients(self, beta_ij: Array) -> Array:
        """Coefficients of the ordinary differential equations.

        Parameters
        ----------
        beta_ij: array
            (`n_pipes`, `n_pipes`,) array of thermal conductance
            coefficients.

        Returns
        -------
        array
            (`n_pipes`, `n_pipes`,) array of coefficients of the ordinary
            differential equations.

        """
        diag_indices = jnp.diag_indices_from(beta_ij)
        ode_coefficients = beta_ij.at[diag_indices].set(
            -beta_ij.sum(axis=1)
        )
        ode_coefficients = ode_coefficients.at[self._n_pipes_over_two:, :].multiply(-1)
        return ode_coefficients

    def _outlet_fluid_temperature_a_in(self, beta_ij: Array) -> float:
        """Inlet coefficient to evaluate the outlet fluid temperature.

        Parameters
        ----------
        beta_ij: array
            (`n_pipes`, `n_pipes`,) array of thermal conductance
            coefficients.

        Returns
        -------
        float
            Coefficient for the inlet fluid temperature.

        """
        # General solution at the bottom-end
        E_0 = self._general_solution_a_0(1., beta_ij)
        # Coefficients for connectivity at the top of the borehole
        # T_f_d(-1) = c_in * T_f_in + c_u @ T_f_u(-1)
        c_in = self._top_connectivity_in
        c_u = self._top_connectivity_u
        # Coefficient for mixing of the fluid at the outlet
        # T_f_out = m_u @ T_f_u(-1)
        m_u = self._mixing_u

        # Intermediate coefficients
        delta_E_0_u = (
            E_0[:self._n_pipes_over_two]
            - E_0[self._n_pipes_over_two:]
        )
        delta_E_0_u = (
            delta_E_0_u[:, :self._n_pipes_over_two] @ c_u
            + delta_E_0_u[:, self._n_pipes_over_two:]
        )
        delta_E_0_in = (
            E_0[self._n_pipes_over_two:]
            - E_0[:self._n_pipes_over_two]
        )
        delta_E_0_in = delta_E_0_in[:, :self._n_pipes_over_two] @ c_in

        # Inlet coefficient to evaluate the outlet fluid temperature
        a_in = m_u @ jnp.linalg.solve(delta_E_0_u, delta_E_0_in)
        return a_in

    def _outlet_fluid_temperature_a_b(self, beta_ij: Array) -> Array:
        """Borehole coefficient to evaluate the outlet fluid temperature.

        Parameters
        ----------
        beta_ij: array
            (`n_pipes`, `n_pipes`,) array of thermal conductance
            coefficients.

        Returns
        -------
        array
            (`n_nodes`,) array of coefficients for the borehole wall
            temperature.

        """
        # General solution at the bottom-end
        E_0 = self._general_solution_a_0(1., beta_ij)
        E_b = self._general_solution_a_b(1., beta_ij)
        # Coefficients for connectivity at the top of the borehole
        # T_f_d(-1) = c_in * T_f_in + c_u @ T_f_u(-1)
        c_in = self._top_connectivity_in
        c_u = self._top_connectivity_u
        # Coefficient for mixing of the fluid at the outlet
        # T_f_out = m_u @ T_f_u(-1)
        m_u = self._mixing_u

        # Intermediate coefficients
        delta_E_0_u = (
            E_0[:self._n_pipes_over_two]
            - E_0[self._n_pipes_over_two:]
        )
        delta_E_0_u = (
            delta_E_0_u[:, :self._n_pipes_over_two] @ c_u
            + delta_E_0_u[:, self._n_pipes_over_two:]
        )
        delta_E_b = (
            E_b[self._n_pipes_over_two:]
            - E_b[:self._n_pipes_over_two]
        )

        # Borehole coefficient to evaluate the outlet fluid temperature
        a_b = m_u @ jnp.linalg.solve(delta_E_0_u, delta_E_b)
        return a_b



# -*- coding: utf-8 -*-
from collections.abc import Callable
from functools import partial
from typing import List, Self, Tuple

from jax import numpy as jnp
from jax import Array, jit, vmap
from jax.typing import ArrayLike

from ..basis import Basis
from .borefield import Borefield
from ..borehole import SingleUTube
from ..path import Path


class Network(Borefield):
    """Borefield with pipes.

    Parameters
    ----------
    boreholes : list of single_u_tube
        Boreholes in the borefield.

    Attributes
    ----------
    n_boreholes : int
        Number of boreholes.
    n_nodes : int
        Number of nodes per borehole.
    L : array
        Borehole lengths (in meters).
    xi : array
        Coordinates of the nodes. They are the same for all boreholes.
    p : array
        (`n_boreholes`, `n_nodes`, 3,) array of node positions.
    dp_dxi : array
        (`n_boreholes`, `n_nodes`, 3,) array of the derivatives of the
        position at the node coordinates.
    J : array
        (`n_boreholes`, `n_nodes`,) array of the norm of the Jacobian at
        the node coordinates.
    s : array
        (`n_boreholes`, `n_nodes`,) array of the longitudinal position at
        the node coordinates.
    w : array
        (`n_boreholes`, `n_nodes`,) array of quadrature weights at the
        node coordinates. These quadrature weights take into account the
        norm of the Jacobian.

    """

    def __init__(self, boreholes: List[SingleUTube]):
        super().__init__(boreholes)

    @partial(jit, static_argnames=['self'])
    def effective_borefield_thermal_resistance(self, m_flow: float | Array, cp_f: float) -> float:
        """Effective borefield thermal resistance.

        Parameters
        ----------
        m_flow : float or array
            Total fluid mass flow rate (in kg/s), or (`n_boreholes`,)
            array of fluid mass flow rate per borehole.
        cp_f : float
            Fluid specific isobaric heat capacity (in J/kg-K).

        Returns
        -------
        float
            Effective borefield thermal resistance (in m-K/W).

        """
        a = jnp.average(
            self._outlet_fluid_temperature(m_flow, cp_f)[0],
            weights=jnp.broadcast_to(
                m_flow, self.n_boreholes
            )
        )
        b = jnp.sum(self._heat_extraction_rate(self.xi, m_flow, cp_f)[0] * self.w)
        # Effective borehole thermal resistance
        R_field = -0.5 * self.L.sum() * (1. + a) / b
        return R_field

    @partial(jit, static_argnames=['self'])
    def g(self, xi: Array | float, m_flow: float | Array, cp_f: float) -> Array:
        """Coefficients to evaluate the heat extraction rate.

        Parameters
        ----------
        xi : array or float
            (M,) array of coordinates along the boreholes.
        m_flow : float or array
            Total fluid mass flow rate (in kg/s), or (`n_boreholes`,)
            array of fluid mass flow rate per borehole.
        cp_f : float
            Fluid specific isobaric heat capacity (in J/kg-K).

        Returns
        -------
        a_in : array or float
            (`n_boreholes`, `M,) array of coefficients for the inlet fluid
            temperature.
        a_b : array
            (`n_boreholes`, M, `n_nodes`,) array of coefficients for the
            borehole wall temperatures.

        """
        return self._heat_extraction_rate(xi, m_flow, cp_f)

    @partial(jit, static_argnames=['self'])
    def g_to_self(self, m_flow: float | Array, cp_f: float) -> Array:
        """Coefficients to evaluate the heat extraction rate at nodes.

        Parameters
        ----------
        m_flow : float or array
            Total fluid mass flow rate (in kg/s), or (`n_boreholes`,)
            array of fluid mass flow rate per borehole.
        cp_f : float
            Fluid specific isobaric heat capacity (in J/kg-K).

        Returns
        -------
        a_in : array or float
            (`n_boreholes, `n_nodes`,) array of coefficients for the inlet
            fluid temperature.
        a_b : array
            (`n_boreholes`, `n_nodes`, `n_nodes`,) array of coefficients
            for the borehole wall temperatures.

        """
        return self._heat_extraction_rate(self.xi, m_flow, cp_f)

    @partial(jit, static_argnames=['self'])
    def fluid_temperature(self, xi: Array | float, T_f_in: float, T_b: Array, m_flow: float | Array, cp_f: float) -> Array:
        """Fluid temperatures.

        Parameters
        ----------
        xi : array or float
            (M,) array of coordinates along the boreholes.
        T_f_in : float
            Inlet fluid temperature (in degree Celsius).
        T_b : array
            (`n_boreholes`, `n_nodes`,) array of borehole wall
            temperatures (in degree Celsius).
        m_flow : float or array
            Total fluid mass flow rate (in kg/s), or (`n_boreholes`,)
            array of fluid mass flow rate per borehole.
        cp_f : float
            Fluid specific isobaric heat capacity (in J/kg-K).

        Returns
        -------
        array
            (`n_boreholes`, M, 2,) array of fluid temperatures (in degree
            Celsius).

        """
        a_in, a_b = self._fluid_temperature(xi, m_flow, cp_f)
        T_f = a_in * T_f_in + vmap(jnp.dot, in_axes=(0, 0), out_axes=0)(a_b, T_b)
        return T_f

    @partial(jit, static_argnames=['self'])
    def heat_extraction_rate(self, xi: Array | float, T_f_in: float, T_b: Array, m_flow: float | Array, cp_f: float) -> Array:
        """Heat extraction rate.

        Parameters
        ----------
        xi : array or float
            (M,) array of coordinates along the boreholes.
        T_f_in : float
            Inlet fluid temperature (in degree Celsius).
        T_b : array
            (`n_boreholes`, `n_nodes`,) array of borehole wall
            temperatures (in degree Celsius).
        m_flow : float or array
            Total fluid mass flow rate (in kg/s), or (`n_boreholes`,)
            array of fluid mass flow rate per borehole.
        cp_f : float
            Fluid specific isobaric heat capacity (in J/kg-K).

        Returns
        -------
        array
            (`n_boreholes`, M,) array of heat extraction rate (in W/m). If
            `xi` is a ``float``, then ``M=0``.

        """
        a_in, a_b = self._heat_extraction_rate(xi, m_flow, cp_f)
        q = a_in * T_f_in + vmap(jnp.dot, in_axes=(0, 0), out_axes=0)(a_b, T_b)
        return q

    @partial(jit, static_argnames=['self'])
    def heat_extraction_rate_to_self(self, T_f_in: float, T_b: Array, m_flow: float | Array, cp_f: float) -> Array:
        """Heat extraction rate at nodes.

        Parameters
        ----------
        T_f_in : float
            Inlet fluid temperature (in degree Celsius).
        T_b : array
            (`n_boreholes`, `n_nodes`,) array of borehole wall
            temperatures (in degree Celsius).
        m_flow : float or array
            Total fluid mass flow rate (in kg/s), or (`n_boreholes`,)
            array of fluid mass flow rate per borehole.
        cp_f : float
            Fluid specific isobaric heat capacity (in J/kg-K).

        Returns
        -------
        array
            (`n_boreholes`, `n_nodes`,) array of heat extraction rate
            (in W/m). If `xi` is a ``float``, then ``M=0``.

        """
        return self.heat_extraction_rate(self.xi, T_f_in, T_b, m_flow, cp_f)

    @partial(jit, static_argnames=['self'])
    def outlet_fluid_temperature(self, T_f_in: float, T_b: Array, m_flow: float | Array, cp_f: float) -> Array:
        """Outlet fluid temperatures.

        Parameters
        ----------
        T_f_in : float
            Inlet fluid temperature (in degree Celsius).
        T_b : array
            (`n_boreholes, `n_nodes`,) array of borehole wall temperatures
            (in degree Celsius).
        m_flow : float or array
            Total fluid mass flow rate (in kg/s), or (`n_boreholes`,)
            array of fluid mass flow rate per borehole.
        cp_f : float
            Fluid specific isobaric heat capacity (in J/kg-K).

        Returns
        -------
        array
            (`n_boreholes`,) array of outlet fluid temperatures (in degree
            Celsius).

        """
        a_in, a_b = self._outlet_fluid_temperature(m_flow, cp_f)
        T_f_out = a_in * T_f_in + vmap(jnp.dot, in_axes=(0, 0), out_axes=0)(a_b, T_b)
        return T_f_out

    def _fluid_temperature(self, xi: Array | float, m_flow: float | Array, cp_f: float) -> Tuple[Array, Array]:
        """Coefficients to evaluate the fluid temperatures.

        Parameters
        ----------
        xi : array or float
            (M,) array of coordinates along the boreholes.
        m_flow : float or array
            Total fluid mass flow rate (in kg/s), or (`n_boreholes`,)
            array of fluid mass flow rate per borehole.
        cp_f : float
            Fluid specific isobaric heat capacity (in J/kg-K).

        Returns
        -------
        a_in : array
            (`n_boreholes`, M, 2,) array of coefficients for the inlet
            fluid temperature.
        a_b : array
            (`n_boreholes`, M, 2, `n_nodes`,) array of coefficients for
            the borehole wall temperature.

        """
        if len(jnp.shape(m_flow)) == 0 or jnp.shape(m_flow)[0] == 1:
            m_flow = jnp.broadcast_to(
                m_flow / self.n_boreholes,
                self.n_boreholes
            )
        a_in, a_b = zip(*[
            borehole._fluid_temperature(xi, _m_flow, cp_f)
            for borehole, _m_flow in zip(self.boreholes, m_flow)
        ])
        a_in = jnp.stack(a_in, axis=0)
        a_b = jnp.stack(a_b, axis=0)
        return a_in, a_b

    def _heat_extraction_rate(self, xi: Array | float, m_flow: float | Array, cp_f: float) -> Tuple[Array, Array]:
        """Coefficients to evaluate the heat extraction rate.

        Parameters
        ----------
        xi : array or float
            (M,) array of coordinates along the boreholes.
        m_flow : float or array
            Total fluid mass flow rate (in kg/s), or (`n_boreholes`,)
            array of fluid mass flow rate per borehole.
        cp_f : float
            Fluid specific isobaric heat capacity (in J/kg-K).

        Returns
        -------
        a_in : array or float
            (`n_boreholes`, M,) array of coefficients for the inlet fluid
            temperature.
        a_b : array
            (`n_boreholes`, M, `n_nodes`,) array of coefficients for the
            borehole wall temperature.

        """
        if len(jnp.shape(m_flow)) == 0 or jnp.shape(m_flow)[0] == 1:
            m_flow = jnp.broadcast_to(
                m_flow / self.n_boreholes,
                self.n_boreholes
            )
        a_in, a_b = zip(*[
            borehole._heat_extraction_rate(xi, _m_flow, cp_f)
            for borehole, _m_flow in zip(self.boreholes, m_flow)
        ])
        a_in = jnp.stack(a_in, axis=0)
        a_b = jnp.stack(a_b, axis=0)
        return a_in, a_b

    def _outlet_fluid_temperature(self, m_flow: float | Array, cp_f: float) -> Tuple[Array, Array]:
        """Coefficients to evaluate the outlet fluid temperatures.

        Parameters
        ----------
        m_flow : float or array
            Total fluid mass flow rate (in kg/s), or (`n_boreholes`,)
            array of fluid mass flow rate per borehole.
        cp_f : float
            Fluid specific isobaric heat capacity (in J/kg-K).

        Returns
        -------
        a_in : array
            (`n_boreholes`,) array of coefficients for the inlet fluid
            temperature.
        a_b : array
            (`n_boreholes`, `n_nodes`,) array of coefficients for the
            borehole wall temperature.

        """
        if len(jnp.shape(m_flow)) == 0 or jnp.shape(m_flow)[0] == 1:
            m_flow = jnp.broadcast_to(
                m_flow / self.n_boreholes,
                self.n_boreholes
            )
        a_in, a_b = zip(*[
            borehole._outlet_fluid_temperature(_m_flow, cp_f)
            for borehole, _m_flow in zip(self.boreholes, m_flow)
        ])
        a_in = jnp.stack(a_in, axis=0)
        a_b = jnp.stack(a_b, axis=0)
        return a_in, a_b

    @classmethod
    def from_positions(cls, L: ArrayLike, D: ArrayLike, r_b: ArrayLike, x: ArrayLike, y: ArrayLike, R_d: ArrayLike | Callable[[float], Array], basis: Basis, n_segments: int, tilt: float = 0., orientation: float = 0., segment_ratios: ArrayLike | None = None, order: int | None = None) -> Self:
        """Field of straight boreholes from their dimensions.

        Parameters
        ----------
        L : array_like
            Borehole length (in meters).
        D : array_like
            Borehole buried depth (in meters).
        r_b : array_like
            Borehole radius (in meters).
        x, y : array_like
            Horizontal position (in meters) of the top end of the
            borehole.
        R_d : array_like or callable
            (2, 2,) array of thermal resistances (in m-K/W), or callable
            that takes the mass flow rate as input (in kg/s) and returns a
            (2, 2,) array.
        basis : basis
            Basis functions.
        n_segments : int
            Number of segments per borehole.
        tilt : array_like, default: ``0.``
            Tilt angle (in radians) of the boreholes with respect to
            vertical.
        orientation : array_like, default: ``0.``
            Orientation (in radians) of the inclined boreholes. An
            inclination toward the x-axis corresponds to an orientation
            of zero.
        segment_ratios : array_like or None, default: None
            Normalized size of the segments. Should total ``1``
            (i.e. ``sum(segment_ratios) = 1``). If `segment_ratios` is
            ``None``, segments of equal size are considered (i.e.
            ``segment_ratios[v] = 1 / n_segments``).

        Returns
        -------
        borefield
            Instance of the `Network` class.

        """
        # Runtime type validation
        if not isinstance(x, ArrayLike):
            raise TypeError(f"Expected arraylike input; got {x}")
        if not isinstance(y, ArrayLike):
            raise TypeError(f"Expected arraylike input; got {y}")
        # Convert input to jax.Array
        x = jnp.asarray(x)
        y = jnp.asarray(y)

        n_boreholes = len(x)
        L = jnp.broadcast_to(L, n_boreholes)
        D = jnp.broadcast_to(D, n_boreholes)
        r_b = jnp.broadcast_to(r_b, n_boreholes)
        tilt = jnp.broadcast_to(tilt, n_boreholes)
        orientation = jnp.broadcast_to(orientation, n_boreholes)
        boreholes = []
        xi = jnp.array([-1., 1.])
        for j in range(n_boreholes):
            p = jnp.array(
                [
                    [x[j], y[j], -D[j]],
                    [x[j] + L[j] * jnp.sin(tilt[j]) * jnp.cos(orientation[j]), y[j] + L[j] * jnp.sin(tilt[j]) * jnp.sin(orientation[j]), -D[j] - L[j] * jnp.cos(tilt[j])],
                ]
            )
            path = Path(xi, p, order=order)
            boreholes.append(SingleUTube(R_d, r_b[j], path, basis, n_segments, segment_ratios=segment_ratios))
        return cls(boreholes)

    @classmethod
    def rectangle_field(cls, N_1: int, N_2: int, B_1: float, B_2: float, L: float, D: float, r_b: float, R_d: ArrayLike | Callable[[float], Array], basis: Basis, n_segments: int, segment_ratios: ArrayLike | None = None, order: int | None = None) -> Self:
        """Field of vertical boreholes in a rectangular configuration.

        Parameters
        ----------
        N_1, N_2 : int
            Number of columns and rows in the borefield.
        B_1, B_2 : float
            Spacing between columns and rows (in meters).
        L : float
            Borehole length (in meters).
        D : float
            Borehole buried depth (in meters).
        r_b : float
            Borehole radius (in meters).
        R_d : array_like or callable
            (2, 2,) array of thermal resistances (in m-K/W), or callable
            that takes the mass flow rate as input (in kg/s) and returns a
            (2, 2,) array.
        basis : basis
            Basis functions.
        n_segments : int
            Number of segments per borehole.
        segment_ratios : array_like or None, default: None
            Normalized size of the segments. Should total ``1``
            (i.e. ``sum(segment_ratios) = 1``). If `segment_ratios` is
            ``None``, segments of equal size are considered (i.e.
            ``segment_ratios[v] = 1 / n_segments``).

        Returns
        -------
        borefield
            Instance of the `Network` class.

        """
        # Borehole positions and orientation
        x = jnp.tile(jnp.arange(N_1), N_2) * B_1
        y = jnp.repeat(jnp.arange(N_2), N_1) * B_2
        return cls.from_positions(L, D, r_b, x, y, R_d, basis, n_segments, segment_ratios=segment_ratios)
        



# -*- coding: utf-8 -*-
from typing import List, Tuple

from jax import numpy as jnp
from matplotlib import pyplot as plt
from matplotlib.axes import Axes
from matplotlib.figure import Figure
from mpl_toolkits.mplot3d.axes3d import Axes3D

from ._format_axis import _format_axis
from ..path import Path


def plot_path(path: Path | List[Path], num: int = 101, view: str = 'all', ax: Axes | Axes3D | None = None) -> Axes | Axes3D | Tuple[Figure, Axes] | Tuple[Figure, Axes3D] | Tuple[Figure, List[Axes | Axes3D]]:
    """Plot path trajectory

    Parameters
    ----------
    path : path or list of path
        Path to be plotted.
    num : int, default: 101
        Number of points along the path.
    view : {'all', 'xy', 'yz', 'xz', '3d'}, default: 'all'
        The view to be plotted. 'all' plots all views on a 2 by 2
        figure.
    ax : axes, axes_3d or None, default: ``None``
        The axis on which to draw the trajectories. Is `ax` is ``None``,
        a new figure and new axes are created.

    Returns
    -------
    fig : figure
        The figure. Only return if `ax` is ``None``.
    ax : axes, axes_3d or list
        The axis.

    """
    if view == 'xy' or view == 'yx':
        return _plot_xy(path, num=num, ax=ax)
    if view == 'yz' or view == 'zy':
        return _plot_yz(path, num=num, ax=ax)
    if view == 'xz' or view == 'zx':
        return _plot_xz(path, num=num, ax=ax)
    if view == '3d':
        return _plot_3d(path, num=num, ax=ax)
    if view == 'all':
        fig = plt.figure(layout='constrained')
        axs = [
            fig.add_subplot(221),
            fig.add_subplot(222, projection='3d'),
            fig.add_subplot(223),
            fig.add_subplot(224),
        ]
        axs[0] = _plot_xy(path, num=num, ax=axs[0])
        axs[1] = _plot_3d(path, num=num, ax=axs[1])
        axs[2] = _plot_xz(path, num=num, ax=axs[2])
        axs[3] = _plot_yz(path, num=num, ax=axs[3])
        return fig, axs


def _plot_xy(path: Path | List[Path], num: int = 101, ax: Axes | None = None) -> Axes | Tuple[Figure, Axes]:
    """Plot xy view of path trajectory

    Parameters
    ----------
    path : path or list of path
        Path to be plotted.
    num : int, default: 101
        Number of points along the path.
    ax : axes or None, default: ``None``
        The axis on which to draw the trajectories. Is `ax` is ``None``,
        a new figure and new axes are created.

    Returns
    -------
    fig : figure
        The figure. Only return if `ax` is ``None``.
    ax : axes
        The axis.

    """
    ax_is_none = ax is None
    if ax is None:
        fig, ax = plt.subplots(layout='constrained')
    _format_axis(ax, axis_labels=[r'$x$ [m]', r'$y$ [m]'], equal=True)

    xi = jnp.linspace(-1., 1., num=num)
    if isinstance(path, list):
        for _path in path:
            p = _path.f_p(xi)
            ax.plot(p[:, 0], p[:, 1])
        for _path in path:
            p = _path.f_p(xi)
            ax.plot(p[0, 0], p[0, 1], 'k.')
    else:
        p = path.f_p(xi)

        ax.plot(p[:, 0], p[:, 1])
        ax.plot(p[0, 0], p[0, 1], 'k.')

    if ax_is_none:
        return fig, ax
    else:
        return ax


def _plot_yz(path: Path | List[Path], num: int = 101, ax: Axes | None = None) -> Axes | Tuple[Figure, Axes]:
    """Plot yz view of path trajectory

    Parameters
    ----------
    path : path or list of path
        Path to be plotted.
    num : int, default: 101
        Number of points along the path.
    ax : axes or None, default: ``None``
        The axis on which to draw the trajectories. Is `ax` is ``None``,
        a new figure and new axes are created.

    Returns
    -------
    fig : figure
        The figure. Only return if `ax` is ``None``.
    ax : axes
        The axis.

    """
    ax_is_none = ax is None
    if ax is None:
        fig, ax = plt.subplots(layout='constrained')
    _format_axis(ax, axis_labels=[r'$y$ [m]', r'$z$ [m]'], equal=True)

    xi = jnp.linspace(-1., 1., num=num)
    if isinstance(path, list):
        for _path in path:
            p = _path.f_p(xi)
            ax.plot(p[:, 1], p[:, 2])
        for _path in path:
            p = _path.f_p(xi)
            ax.plot(p[0, 1], p[0, 2], 'k.')
    else:
        p = path.f_p(xi)

        ax.plot(p[:, 1], p[:, 2])
        ax.plot(p[0, 1], p[0, 2], 'k.')

    if ax_is_none:
        return fig, ax
    else:
        return ax


def _plot_xz(path: Path | List[Path], num: int = 101, ax: Axes | None = None) -> Axes | Tuple[Figure, Axes]:
    """Plot xz view of path trajectory

    Parameters
    ----------
    path : path or list of path
        Path to be plotted.
    num : int, default: 101
        Number of points along the path.
    ax : axes or None, default: ``None``
        The axis on which to draw the trajectories. Is `ax` is ``None``,
        a new figure and new axes are created.

    Returns
    -------
    fig : figure
        The figure. Only return if `ax` is ``None``.
    ax : axes
        The axis.

    """
    ax_is_none = ax is None
    if ax is None:
        fig, ax = plt.subplots(layout='constrained')
    _format_axis(ax, axis_labels=[r'$x$ [m]', r'$z$ [m]'], equal=True)

    xi = jnp.linspace(-1., 1., num=num)
    if isinstance(path, list):
        for _path in path:
            p = _path.f_p(xi)
            ax.plot(p[:, 0], p[:, 2])
        for _path in path:
            p = _path.f_p(xi)
            ax.plot(p[0, 0], p[0, 2], 'k.')
    else:
        p = path.f_p(xi)

        ax.plot(p[:, 0], p[:, 2])
        ax.plot(p[0, 0], p[0, 2], 'k.')

    if ax_is_none:
        return fig, ax
    else:
        return ax


def _plot_3d(path: Path | List[Path], num: int = 101, ax: Axes3D | None = None) -> Axes3D | Tuple[Figure, Axes3D]:
    """Plot 3d view of path trajectory

    Parameters
    ----------
    path : path or list of path
        Path to be plotted.
    num : int, default: 101
        Number of points along the path.
    ax : axes_3d or None, default: ``None``
        The axis on which to draw the trajectories. Is `ax` is ``None``,
        a new figure and new axes are created.

    Returns
    -------
    fig : figure
        The figure. Only return if `ax` is ``None``.
    ax : axes_3d
        The axis.

    """
    ax_is_none = ax is None
    if ax is None:
        fig, ax = plt.subplots(subplot_kw={'projection': '3d'}, layout='constrained')
    _format_axis(ax, axis_labels=[r'$x$ [m]', r'$y$ [m]', r'$z$ [m]'], equal=True)

    xi = jnp.linspace(-1., 1., num=num)
    if isinstance(path, list):
        for _path in path:
            p = _path.f_p(xi)
            ax.plot(p[:, 0], p[:, 1], p[:, 2])
        for _path in path:
            p = _path.f_p(xi)
            ax.plot(p[0, 0], p[0, 1], p[0, 2], 'k.')
    else:
        p = path.f_p(xi)

        ax.plot(p[:, 0], p[:, 1], p[:, 2])
        ax.plot(p[0, 0], p[0, 1], p[0, 2], 'k.')

    if ax_is_none:
        return fig, ax
    else:
        return ax



# -*- coding: utf-8 -*-
from jax import numpy as jnp


def conduction_thermal_resistance_circular_pipe(r_p_in: float, r_p_out: float, k_p: float) -> float:
    """Conduction thermal resistance in circular pipe.

    Parameters
    ----------
    r_p_in : float
        Inner radius of the pipe (in meters).
    r_p_out : float
        Outer radius of the pipe (in meters).
    k_p : float
        Pipe thermal conductivity (in W/m-K).

    Returns
    -------
    R_p : float
        Conduction thermal resistance (in m-K/W).

    """
    R_p = jnp.log(r_p_out / r_p_in) / (2 * jnp.pi * k_p)
    return R_p



# -*- coding: utf-8 -*-
from jax import numpy as jnp


def reynolds_number_annular_pipe(m_flow: float, r_p_in: float, r_p_out: float, mu_f: float, rho_f: float) -> float:
    """Reynolds number in an annular pipe.

    Parameters
    ----------
    m_flow : float
        Fluid mass flow rate (in kg/s).
    r_p_in : float
        Outer radius of the inner pipe (in meters).
    r_p_out : float
        Inner radius of the outer pipe (in meters).
    mu_f : float
        Fluid dynamic viscosity (in kg/m-s).
    rho_f : float
        Fluid density (in kg/m3).

    Returns
    -------
    float
        Reynolds number.

    """
    # Hydraulic diameter
    D = 2 * (r_p_out - r_p_in)
    # Fluid velocity
    V_flow = jnp.abs(m_flow) / rho_f
    A_cs = jnp.pi * (r_p_out**2 - r_p_in**2)
    V = V_flow / A_cs
    # Reynolds number
    Re = rho_f * V * D / mu_f
    return Re


def reynolds_number_circular_pipe(m_flow: float, r_p: float, mu_f: float, rho_f: float) -> float:
    """Reynolds number in a circular pipe.

    Parameters
    ----------
    m_flow : float
        Fluid mass flow rate (in kg/s).
    r_p : float
        Inner radius of the pipe (in meters).
    mu_f : float
        Fluid dynamic viscosity (in kg/m-s).
    rho_f : float
        Fluid density (in kg/m3).

    Returns
    -------
    float
        Reynolds number.

    """
    # Hydraulic diameter
    D = 2 * r_p
    # Fluid velocity
    V_flow = jnp.abs(m_flow) / rho_f
    A_cs = jnp.pi * r_p**2
    V = V_flow / A_cs
    # Reynolds number
    Re = rho_f * V * D / mu_f
    return Re
